#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

r"""
Acquisition function for predictive entropy search for multi-objective Bayesian
optimization (PES). The code does not support constraint handling.

NOTE: The PES acquisition might not be differentiable. As a result, we recommend
optimizing the acquisition function using finite differences.

References:

.. [Garrido-Merchan2019]
    E. Garrido-Merchan and D. Hernandez-Lobato. Predictive Entropy Search for
    Multi-objective Bayesian Optimization with Constraints. Neurocomputing. 2019.
    The computation follows the procedure described in the supplementary material:
    https://www.sciencedirect.com/science/article/abs/pii/S0925231219308525

"""

from __future__ import annotations

import torch
from botorch.acquisition.acquisition import AcquisitionFunction
from botorch.exceptions import InputDataError
from botorch.exceptions.errors import UnsupportedError
from botorch.models.model import Model
from botorch.models.model_list_gp_regression import ModelListGP
from botorch.models.utils import check_no_nans
from botorch.utils.transforms import (
    average_over_ensemble_models,
    concatenate_pending_points,
    t_batch_mode_transform,
)
from torch import Tensor
from torch.distributions import Normal


class qMultiObjectivePredictiveEntropySearch(AcquisitionFunction):
    r"""The acquisition function for Predictive Entropy Search. The code supports
    both single and multiple objectives as well as batching.

    This acquisition function approximates the mutual information between the
    observation at a candidate point `X` and the Pareto optimal input using the
    moment-matching procedure known as expectation propagation (EP).

    See the Appendix of [Garrido-Merchan2019]_ for the description of the EP
    procedure.

    IMPORTANT NOTES:
    (i) The PES acquisition function estimated using EP is sometimes not
    differentiable, and therefore we advise using a finite-difference estimate of
    the gradient as opposed to the gradients identified using automatic
    differentiation, which occasionally outputs `nan` values.

    The source of this differentiability is in the `_update_damping` function, which
    finds the damping factor `a` that is used to update the EP parameters
    `a * param_new + (1 - a) * param_old`. The damping factor has to ensure
    that the updated covariance matrices, `a * cov_f_new + (1 - a) cov_f_old`, is
    positive semi-definiteness. We follow the original paper, which identifies
    `a` via a successive halving scheme i.e. we check `a=1` then `a=0.5` etc. This
    procedure means `a` is a function of the test input `X`. This function is not
    differentiable  in `X`.

    (ii) EP could potentially fail for a number of reasons:

        (a) When the sampled Pareto optimal points `x_p` is poor compared to the
        training or testing data `x_n`.

        (b) When the training or testing data `x_n` is close the Pareto optimal
        points `x_p`.

        (c) When the convergence threshold is set too small.


        Problem (a) occurs because we have to compute the variable:
        `alpha = (mean(x_n) - mean(x_p)) / std(x_n - x_p)`, which becomes very
        large when `x_n` is better than `x_p` with high-probability. This leads to a
        log(0) error when we compute `log(1 - cdf(alpha))`. We have preemptively
        clamped some values depending on `1`alpha` in order to mitigate this.

        Problem (b) occurs because we have to compute matrix inverses for the
        two-dimensional marginals (x_n, x_p). To address this we manually add jitter
        to the diagonal of the covariance matrix i.e. `ep_jitter` when training and
        `test_jitter` when testing. The default choice is not always appropriate
        because the same jitter is used for the inversion of the covariance
        and precision matrix, which are on different scales.

        TODO: come up with strategy to adaptively update the jitter.

        Problem (c) occurs because a smaller threshold usually means that more EP
        iterations are required. Running too many EP iterations could lead to
        invertibility problems such as in problem (b). Setting a larger threshold
        or reducing the number of EP iterations could alleviate this.

    (iii) The estimated acquisition value could be negative.
    """

    def __init__(
        self,
        model: Model,
        pareto_sets: Tensor,
        maximize: bool = True,
        X_pending: Tensor | None = None,
        max_ep_iterations: int = 250,
        ep_jitter: float = 1e-4,
        test_jitter: float = 1e-4,
        threshold: float = 1e-2,
    ) -> None:
        r"""Multi-objective predictive entropy search acquisition function.

        Args:
            model: A fitted batched model with `M` number of outputs.
            pareto_sets: A `num_pareto_samples x P x d`-dim tensor containing the
                Pareto optimal set of inputs, where `P` is the number of pareto
                optimal points. The points in each sample have to be discrete
                otherwise expectation propagation will fail.
            maximize: If true, we consider a maximization problem.
            X_pending: A `m x d`-dim Tensor of `m` design points that have been
                submitted for function evaluation, but have not yet been evaluated.
            max_ep_iterations: The maximum number of expectation propagation
                iterations. (The minimum number of iterations is set at 3.)
            ep_jitter: The amount of jitter added for the matrix inversion that
                occurs during the expectation propagation update during the training
                phase.
            test_jitter: The amount of jitter added for the matrix inversion that
                occurs during the expectation propagation update in the testing
                phase.
            threshold: The convergence threshold for expectation propagation. This
                assesses the relative change in the mean and covariance. We default
                to one percent change i.e. `threshold = 1e-2`.
        """
        super().__init__(model=model)

        self.model = model
        self.maximize = maximize
        self.set_X_pending(X_pending)

        if model.num_outputs > 1 or isinstance(model, ModelListGP):
            train_X = self.model.train_inputs[0][0]
        else:
            train_X = self.model.train_inputs[0]

        # Batch GP models (e.g. fantasized models) are not currently supported
        if train_X.ndim > 2:
            raise NotImplementedError(
                "Batch GP models (e.g. fantasized models) are not supported."
            )

        if pareto_sets.ndim != 3 or pareto_sets.shape[-1] != train_X.shape[-1]:
            raise UnsupportedError(
                "The Pareto set should have a shape of "
                "`num_pareto_samples x num_pareto_points x input_dim`."
            )
        else:
            self.pareto_sets = pareto_sets

        # add the pareto set to the existing training data
        self.num_pareto_samples = pareto_sets.shape[0]

        self.augmented_X = torch.cat(
            [train_X.repeat(self.num_pareto_samples, 1, 1), self.pareto_sets], dim=-2
        )
        self.max_ep_iterations = max_ep_iterations
        self.ep_jitter = ep_jitter
        self.test_jitter = test_jitter
        self.threshold = threshold
        self._expectation_propagation()

    def _expectation_propagation(self) -> None:
        r"""Perform expectation propagation to obtain the covariance factors that
        depend on the Pareto sets.

        The updates are performed in the natural parameter space. For a multivariate
        normal distribution with mean mu and covariance Sigma, we call Sigma^{-1}
        the natural covariance and Sigma^{-1} mu the natural mean.
        """
        ###########################################################################
        # INITIALIZATION
        ###########################################################################
        M = self.model.num_outputs

        if self.model.num_outputs > 1 or isinstance(self.model, ModelListGP):
            train_X = self.model.train_inputs[0][0]
        else:
            train_X = self.model.train_inputs[0]

        tkwargs = {"dtype": train_X.dtype, "device": train_X.device}
        N = len(train_X)
        num_pareto_samples = self.num_pareto_samples
        P = self.pareto_sets.shape[-2]

        # initialize the predictive natural mean and variances
        (
            pred_nat_mean,
            pred_nat_cov,
            pred_mean,
            pred_cov,
        ) = _initialize_predictive_matrices(
            X=self.augmented_X,
            model=self.model,
            observation_noise=False,
            jitter=self.ep_jitter,
            natural=True,
        )

        pred_f_mean = pred_mean[..., 0:M, :]
        pred_f_nat_mean = pred_nat_mean[..., 0:M, :]
        pred_f_cov = pred_cov[..., 0:M, :, :]
        pred_f_nat_cov = pred_nat_cov[..., 0:M, :, :]

        # initialize the marginals
        # `num_pareto_samples x M x (N + P)`
        mean_f = pred_f_mean.clone()
        nat_mean_f = pred_f_nat_mean.clone()
        # `num_pareto_samples x M x (N + P) x (N + P)`
        cov_f = pred_f_cov.clone()
        nat_cov_f = pred_f_nat_cov.clone()

        # initialize omega the function which encodes the fact that the pareto points
        # are optimal in the feasible space i.e. any point in the feasible space
        # should not dominate the Pareto efficient points.

        # `num_pareto_samples x M x (N + P) x P x 2`
        omega_f_nat_mean = torch.zeros((num_pareto_samples, M, N + P, P, 2), **tkwargs)
        # `num_pareto_samples x M x (N + P) x P x 2 x 2`
        omega_f_nat_cov = torch.zeros(
            (num_pareto_samples, M, N + P, P, 2, 2), **tkwargs
        )

        ###########################################################################
        # EXPECTATION PROPAGATION
        ###########################################################################
        damping = torch.ones(num_pareto_samples, M, **tkwargs)

        iteration = 0
        while (torch.sum(damping) > 0) and (iteration < self.max_ep_iterations):
            # Compute the new natural mean and covariance
            ####################################################################
            # OBJECTIVE FUNCTION: OMEGA UPDATE
            ####################################################################
            omega_f_nat_mean_new, omega_f_nat_cov_new = _safe_update_omega(
                mean_f=mean_f,
                cov_f=cov_f,
                omega_f_nat_mean=omega_f_nat_mean,
                omega_f_nat_cov=omega_f_nat_cov,
                N=N,
                P=P,
                M=M,
                maximize=self.maximize,
                jitter=self.ep_jitter,
            )

            ####################################################################
            # OBJECTIVE FUNCTION: MARGINAL UPDATE
            ####################################################################
            nat_mean_f_new, nat_cov_f_new = _update_marginals(
                pred_f_nat_mean=pred_f_nat_mean,
                pred_f_nat_cov=pred_f_nat_cov,
                omega_f_nat_mean=omega_f_nat_mean_new,
                omega_f_nat_cov=omega_f_nat_cov_new,
                N=N,
                P=P,
            )
            ########################################################################
            # OBJECTIVE FUNCTION: DAMPING UPDATE
            ########################################################################
            # update damping of objectives
            damping, cholesky_nat_cov_f = _update_damping(
                nat_cov=nat_cov_f,
                nat_cov_new=nat_cov_f_new,
                damping_factor=damping,
                jitter=self.ep_jitter,
            )
            check_no_nans(cholesky_nat_cov_f)
            ########################################################################
            # OBJECTIVE FUNCTION: DAMPED UPDATE
            ########################################################################
            # Damp update of omega
            omega_f_nat_mean = _damped_update(
                old_factor=omega_f_nat_mean,
                new_factor=omega_f_nat_mean_new,
                damping_factor=damping,
            )

            omega_f_nat_cov = _damped_update(
                old_factor=omega_f_nat_cov,
                new_factor=omega_f_nat_cov_new,
                damping_factor=damping,
            )
            # update the mean and covariance
            nat_mean_f = _damped_update(
                old_factor=nat_mean_f, new_factor=nat_mean_f_new, damping_factor=damping
            )
            nat_cov_f = _damped_update(
                old_factor=nat_cov_f, new_factor=nat_cov_f_new, damping_factor=damping
            )

            # compute cholesky inverse
            cov_f_new = torch.cholesky_inverse(cholesky_nat_cov_f)
            mean_f_new = torch.einsum("...ij,...j->...i", cov_f_new, nat_mean_f)
            check_no_nans(cov_f_new)
            ########################################################################
            # OBJECTIVE FUNCTION: CONVERGENCE UPDATE
            ########################################################################
            # Set the damping to zero when the change in the mean and
            # covariance is less than the threshold
            damping, delta_mean_f, delta_cov_f = _update_damping_when_converged(
                mean_old=mean_f,
                mean_new=mean_f_new,
                cov_old=cov_f,
                cov_new=cov_f_new,
                damping_factor=damping,
                threshold=self.threshold,
                iteration=iteration,
            )
            cov_f = cov_f_new
            mean_f = mean_f_new
            iteration = iteration + 1

        ############################################################################
        # SAVE OMEGA AND PHI FACTORS
        ############################################################################
        check_no_nans(omega_f_nat_mean)
        check_no_nans(omega_f_nat_cov)
        # save phi and omega for the forward
        self._omega_f_nat_mean = omega_f_nat_mean
        self._omega_f_nat_cov = omega_f_nat_cov

    def _compute_information_gain(self, X: Tensor) -> Tensor:
        r"""Evaluate qMultiObjectivePredictiveEntropySearch on the candidate set `X`.

        Args:
            X: A `batch_shape x q x d`-dim Tensor of t-batches with `q` `d`-dim
                design points each.

        Returns:
            A `batch_shape'`-dim Tensor of Predictive Entropy Search values at the
            given design points `X`.
        """
        tkwargs = {"dtype": X.dtype, "device": X.device}
        batch_shape = X.shape[0:-2]
        q = X.shape[-2]
        M = self.model.num_outputs

        if M > 1 or isinstance(self.model, ModelListGP):
            N = len(self.model.train_inputs[0][0])
        else:
            N = len(self.model.train_inputs[0])
        P = self.pareto_sets.shape[-2]
        num_pareto_samples = self.num_pareto_samples
        ###########################################################################
        # AUGMENT X WITH THE SAMPLED PARETO SET
        ###########################################################################
        new_shape = batch_shape + torch.Size([num_pareto_samples]) + X.shape[-2:]
        expanded_X = X.unsqueeze(-3).expand(new_shape)
        expanded_ps = self.pareto_sets.expand(X.shape[0:-2] + self.pareto_sets.shape)
        # `batch_shape x num_pareto_samples x (q + P) x d`
        aug_X = torch.cat([expanded_X, expanded_ps], dim=-2)

        ###########################################################################
        # COMPUTE THE POSTERIORS AND OBSERVATION NOISE
        ###########################################################################
        # compute predictive distribution without observation noise
        (
            pred_nat_mean,
            pred_nat_cov,
            pred_mean,
            pred_cov,
        ) = _initialize_predictive_matrices(
            X=aug_X,
            model=self.model,
            observation_noise=True,
            jitter=self.test_jitter,
            natural=True,
        )

        pred_f_mean = pred_mean[..., 0:M, :]
        pred_f_nat_mean = pred_nat_mean[..., 0:M, :]
        pred_f_cov = pred_cov[..., 0:M, :, :]
        pred_f_nat_cov = pred_nat_cov[..., 0:M, :, :]

        (_, _, _, pred_cov_noise) = _initialize_predictive_matrices(
            X=aug_X,
            model=self.model,
            observation_noise=True,
            jitter=self.test_jitter,
            natural=False,
        )

        pred_f_cov_noise = pred_cov_noise[..., 0:M, :, :]
        observation_noise = pred_f_cov_noise - pred_f_cov
        ###########################################################################
        # INITIALIZE THE EP FACTORS
        ###########################################################################
        # `batch_shape x num_pareto_samples x M x (q + P) x P x 2`
        omega_f_nat_mean = torch.zeros(
            batch_shape + torch.Size([num_pareto_samples, M, q + P, P, 2]), **tkwargs
        )
        # `batch_shape x num_pareto_samples x M x (q + P) x P x 2 x 2`
        omega_f_nat_cov = torch.zeros(
            batch_shape + torch.Size([num_pareto_samples, M, q + P, P, 2, 2]), **tkwargs
        )
        ###########################################################################
        # RUN EP ONCE
        ###########################################################################
        # run update omega once
        omega_f_nat_mean, omega_f_nat_cov = _safe_update_omega(
            mean_f=pred_f_mean,
            cov_f=pred_f_cov,
            omega_f_nat_mean=omega_f_nat_mean,
            omega_f_nat_cov=omega_f_nat_cov,
            N=q,
            P=P,
            M=M,
            maximize=self.maximize,
            jitter=self.test_jitter,
        )
        ###########################################################################
        # ADD THE CACHE FACTORS BACK
        ###########################################################################
        omega_f_nat_mean, omega_f_nat_cov = _augment_factors_with_cached_factors(
            q=q,
            N=N,
            omega_f_nat_mean=omega_f_nat_mean,
            cached_omega_f_nat_mean=self._omega_f_nat_mean,
            omega_f_nat_cov=omega_f_nat_cov,
            cached_omega_f_nat_cov=self._omega_f_nat_cov,
        )
        ###########################################################################
        # COMPUTE THE MARGINAL
        ###########################################################################
        nat_mean_f, nat_cov_f = _update_marginals(
            pred_f_nat_mean=pred_f_nat_mean,
            pred_f_nat_cov=pred_f_nat_cov,
            omega_f_nat_mean=omega_f_nat_mean,
            omega_f_nat_cov=omega_f_nat_cov,
            N=q,
            P=P,
        )
        ###########################################################################
        # COMPUTE THE DAMPED UPDATE
        ###########################################################################
        # # update damping of objectives
        damping = torch.ones(
            batch_shape + torch.Size([num_pareto_samples, M]), **tkwargs
        )
        damping, cholesky_nat_cov_f_new = _update_damping(
            nat_cov=pred_f_nat_cov,
            nat_cov_new=nat_cov_f,
            damping_factor=damping,
            jitter=self.test_jitter,
        )

        # invert matrix
        cov_f_new = torch.cholesky_inverse(cholesky_nat_cov_f_new)
        check_no_nans(cov_f_new)

        ###########################################################################
        # COMPUTE THE LOG DETERMINANTS
        ###########################################################################
        # compute the initial log determinant term
        log_det_pred_f_cov_noise = _compute_log_determinant(cov=pred_f_cov_noise, q=q)
        # compute the post log determinant term
        log_det_cov_f = _compute_log_determinant(cov=cov_f_new + observation_noise, q=q)

        ###########################################################################
        # COMPUTE THE ACQUISITION FUNCTION
        ###########################################################################
        q_pes_f = log_det_pred_f_cov_noise - log_det_cov_f
        check_no_nans(q_pes_f)

        return 0.5 * q_pes_f

    @concatenate_pending_points
    @t_batch_mode_transform()
    @average_over_ensemble_models
    def forward(self, X: Tensor) -> Tensor:
        r"""Evaluate qMultiObjectivePredictiveEntropySearch on the candidate set `X`.

        Args:
            X: A `batch_shape x q x d`-dim Tensor of t-batches with `q` `d`-dim
                design points each.

        Returns:
            A `batch_shape'`-dim Tensor of acquisition values at the given design
            points `X`.
        """
        return self._compute_information_gain(X)


def log_cdf_robust(x: Tensor) -> Tensor:
    r"""Computes the logarithm of the normal cumulative density robustly. This uses
    the approximation log(1-z) ~ -z when z is small:

    if x > 5:
        log(cdf(x)) = log(1-cdf(-x)) approx -cdf(-x)
    else:
        log(cdf(x)).

    Args:
        x: a `x_shape`-dim Tensor.

    Returns
        A `x_shape`-dim Tensor.
    """
    CLAMP_LB = torch.finfo(x.dtype).eps
    NEG_INF = torch.finfo(x.dtype).min
    normal = Normal(torch.zeros_like(x), torch.ones_like(x))
    cdf_x = normal.cdf(x)
    neg_cdf_neg_x = -normal.cdf(-x)
    log_cdf_x = torch.where(x < 5, torch.log(cdf_x), neg_cdf_neg_x)

    return log_cdf_x.clamp(NEG_INF, -CLAMP_LB)


def _initialize_predictive_matrices(
    X: Tensor,
    model: Model,
    observation_noise: bool = True,
    jitter: float = 1e-4,
    natural: bool = True,
) -> tuple[Tensor, Tensor, Tensor, Tensor]:
    r"""Initializes the natural predictive mean and covariance matrix. For a
    multivariate normal distribution with mean mu and covariance Sigma, the natural
    mean is Sigma^{-1} mu and the natural covariance is Sigma^{-1}.

    Args:
        X: A `batch_shape x R x d`-dim Tensor.
        model: The fitted model.
        observation_noise: If true, the posterior is computed with observation noise.
        jitter: The jitter added to the covariance matrix.
        natural: If true, we compute the natural statistics as well.

    Return:
        A four-element tuple containing

        - pred_nat_mean: A `batch_shape x num_outputs x R `-dim Tensor containing the
            predictive natural mean vectors.
        - pred_nat_cov: A `batch_shape x num_outputs x R x R`-dim Tensor containing
            the predictive natural covariance matrices.
        - pred_mean: A `batch_shape x num_outputs x R`-dim Tensor containing the
            predictive mean vectors.
        - pred_cov: A `batch_shape x num_outputs x R x R`-dim Tensor containing the
            predictive covariance matrices.
    """
    tkwargs = {"dtype": X.dtype, "device": X.device}
    # compute the predictive mean and covariances at X
    posterior = model.posterior(X, observation_noise=observation_noise)

    # `batch_shape x (R * num_outputs) x (R * num_outputs)`
    init_pred_cov = posterior.mvn.covariance_matrix
    num_outputs = model.num_outputs
    R = int(init_pred_cov.shape[-1] / num_outputs)
    pred_cov = [
        init_pred_cov[..., (m * R) : ((m + 1) * R), (m * R) : ((m + 1) * R)].unsqueeze(
            -1
        )
        for m in range(num_outputs)
    ]
    # `batch_shape x R x R x num_outputs` (before swap axes)
    # `batch_shape x num_outputs x R * R`
    pred_cov = torch.cat(pred_cov, axis=-1).swapaxes(-2, -1).swapaxes(-3, -2)
    identity = torch.diag_embed(torch.ones(pred_cov.shape[:-1], **tkwargs))
    pred_cov = pred_cov + jitter * identity

    # `batch_shape x num_outputs x R`
    pred_mean = posterior.mean.swapaxes(-2, -1)

    #############################################################
    if natural:
        # natural parameters
        # `batch_shape x num_outputs x R x R`
        cholesky_pred_cov, _ = torch.linalg.cholesky_ex(pred_cov)
        pred_nat_cov = torch.cholesky_inverse(cholesky_pred_cov)

        # `batch_shape x num_outputs x R`
        pred_nat_mean = torch.einsum("...ij,...j->...i", pred_nat_cov, pred_mean)

        return pred_nat_mean, pred_nat_cov, pred_mean, pred_cov
    else:
        return None, None, pred_mean, pred_cov


def _get_omega_f_contribution(
    mean: Tensor, cov: Tensor, N: int, P: int, M: int
) -> tuple[Tensor, Tensor]:
    r"""Extract the mean vector and covariance matrix corresponding to the `2 x 2`
    multivariate normal blocks in the objective model between the points in `X` and
    the Pareto optimal set.

    [There is likely a more efficient way to do this.]

    Args:
        mean: A `batch_shape x M x (N + P)`-dim Tensor containing the natural
            mean matrix for the objectives.
        cov: A `batch_shape x M x (N + P) x (N + P)`-dim Tensor containing
            the natural mean matrix for the objectives.
        N: The number of design points.
        P: The number of Pareto optimal points.
        M: The number of objectives.

    Return:
        A two-element tuple containing

        - mean_fX_fS: A `batch_shape x M x (N + P) x P x 2`-dim Tensor containing the
            means of the inputs and Pareto optimal points.
        - cov_fX_fS: A `batch_shape x M x (N + P) x P x 2 x 2`-dim Tensor containing
            the covariances between the inputs and Pareto optimal points.
    """
    tkwargs = {"dtype": mean.dtype, "device": mean.device}
    batch_shape = mean.shape[:-2]
    # `batch_shape x M x (N + P) x P x 2 x 2`
    cov_fX_fS = torch.zeros(batch_shape + torch.Size([M, N + P, P, 2, 2]), **tkwargs)
    # `batch_shape x M x (N + P) x P x 2`
    mean_fX_fS = torch.zeros(batch_shape + torch.Size([M, N + P, P, 2]), **tkwargs)

    # `batch_shape x M x (N + P) x P`
    mean_fX_fS[..., 0] = mean.unsqueeze(-1).expand(mean.shape + torch.Size([P]))
    # `batch_shape x M x (N + P) x P`
    mean_fX_fS[..., 1] = (
        mean[..., N:].unsqueeze(-2).expand(mean.shape + torch.Size([P]))
    )
    # `batch_shape x M x (N + P) x P`
    cov_fX_fS[..., 0, 0] = (
        cov[..., range(N + P), range(N + P)]
        .unsqueeze(-1)
        .expand(batch_shape + torch.Size([M, N + P, P]))
    )
    # `batch_shape x M x (N + P) x P`
    cov_fX_fS[..., 1, 1] = (
        cov[..., range(N, N + P), range(N, N + P)]
        .unsqueeze(-2)
        .expand(batch_shape + torch.Size([M, N + P, P]))
    )

    for p in range(P):
        # `batch_shape x M x (N + P)`
        cov_p = cov[..., range(N + P), N + p]
        cov_fX_fS[..., p, 0, 1] = cov_p
        cov_fX_fS[..., p, 1, 0] = cov_p

    return mean_fX_fS, cov_fX_fS


def _replace_pareto_diagonal(A: Tensor) -> Tensor:
    """Replace the pareto diagonal with identity matricx.

    The Pareto diagonal of the omega factor shouldn't be updated because does not
    contribute anything: `omega(x_p, x_p) = 1` for any pareto optimal input `x_p`.

    Args:
        A: a `batch_shape x M x (N + P) x P x 2 x 2`-dim Tensor.

    Returns:
        A `batch_shape x M x (N + P) x P x 2 x 2`-dim Tensor, where the Pareto
        diagonal is padded with identity matrices.
    """
    tkwargs = {"dtype": A.dtype, "device": A.device}
    batch_shape = A.shape[:-5]
    P = A.shape[-3]
    N = A.shape[-4] - P
    M = A.shape[-5]
    identity = torch.diag_embed(torch.ones(batch_shape + torch.Size([M, 2]), **tkwargs))
    for p in range(P):
        A[..., N + p, p, :, :] = identity

    return A


def _update_omega(
    mean_f: Tensor,
    cov_f: Tensor,
    omega_f_nat_mean: Tensor,
    omega_f_nat_cov: Tensor,
    N: int,
    P: int,
    M: int,
    maximize: bool = True,
    jitter: float = 1e-6,
) -> tuple[Tensor, Tensor]:
    r"""Computes the new omega factors by matching the moments.

    Args:
        mean_f: A `batch_shape x M x (N + P)`-dim Tensor containing the mean vector
            for the objectives.
        cov_f: A `batch_shape x M x (N + P) x (N + P)`-dim Tensor containing the
            covariance matrix for the objectives.
        omega_f_nat_mean: A `batch_shape x M x (N + P) x P x 2`-dim Tensor containing
            the omega natural mean factors for the objective matrix.
        omega_f_nat_cov: A `batch_shape x M x (N + P) x P x 2 x 2`-dim Tensor
            containing the omega natural covariance factors for the objective matrix.
        N: The number of design points.
        M: The number of Pareto optimal points.
        M: The number of objectives.
        maximize: If true, we consider the Pareto maximum domination relation.
        jitter: The jitter for the matrix inverse.

    Return:
        A two-element tuple containing

        - omega_f_nat_mean_new: A `batch_shape x M x (N + P) x P x 2` containing the
            new omega natural mean factors for the objective matrix.
        - omega_f_nat_cov_new: A `batch_shape x M x (N + P) x P x 2 x 2` containing
            the new omega natural covariance factors for the objective matrix.
    """
    tkwargs = {"dtype": mean_f.dtype, "device": mean_f.device}
    CLAMP_LB = torch.finfo(tkwargs["dtype"]).eps
    NEG_INF = torch.finfo(tkwargs["dtype"]).min
    weight = 1.0 if maximize else -1.0
    ###############################################################################
    # EXTRACT THE NECESSARY COMPONENTS
    ###############################################################################
    # `batch_shape x M x (N + P) x P x 2`-dim mean
    # `batch_shape x M x (N + P) x P x 2 x 2`-dim covariance
    mean_fX_fS, cov_fX_fS = _get_omega_f_contribution(mean_f, cov_f, N, P, M)
    identity = torch.diag_embed(torch.ones(cov_fX_fS.shape[:-1], **tkwargs))
    # remove the Pareto diagonal
    cov_fX_fS = _replace_pareto_diagonal(cov_fX_fS + jitter * identity)
    nat_cov_fX_fS = torch.inverse(cov_fX_fS)
    nat_mean_fX_fS = torch.einsum("...ij,...j->...i", nat_cov_fX_fS, mean_fX_fS)

    ###############################################################################
    # COMPUTE THE CAVITIES
    ###############################################################################
    # cavity distribution
    # natural parameters
    cav_nat_mean_f = nat_mean_fX_fS - omega_f_nat_mean
    cav_nat_cov_f = nat_cov_fX_fS - omega_f_nat_cov

    # transform to standard parameters
    # remove the Pareto diagonal
    cav_nat_cov_f = _replace_pareto_diagonal(cav_nat_cov_f)
    identity = torch.diag_embed(torch.ones(cav_nat_cov_f.shape[:-1], **tkwargs))
    cav_cov_f = torch.inverse(cav_nat_cov_f + jitter * identity)

    cav_mean_f = torch.einsum("...ij,...j->...i", cav_cov_f, cav_nat_mean_f)

    ###############################################################################
    # COMPUTE THE NORMALIZATION CONSTANT
    ###############################################################################
    # `batch_shape x M x (N + P) x P`
    # Equation 29
    cav_var_fX_minus_fS = (
        cav_cov_f[..., 0, 0] + cav_cov_f[..., 1, 1] - 2 * cav_cov_f[..., 0, 1]
    ).clamp_min(CLAMP_LB)
    cav_std_fX_minus_fS = torch.sqrt(cav_var_fX_minus_fS).clamp_min(CLAMP_LB)

    # `batch_shape x M x (N + P) x P`
    cav_mean_fX_minus_fS = weight * (cav_mean_f[..., 0] - cav_mean_f[..., 1])

    # Equation 30
    cav_alpha = cav_mean_fX_minus_fS / cav_std_fX_minus_fS
    # compute alpha pdf and cdf
    normal_alpha = Normal(torch.zeros_like(cav_alpha), torch.ones_like(cav_alpha))
    # `batch_shape x M x (N + P) x P`
    cav_alpha_log_cdf = log_cdf_robust(cav_alpha)
    # `batch_shape x M x (N + P) x P`
    cav_alpha_log_pdf = normal_alpha.log_prob(cav_alpha).clamp_min(NEG_INF)
    # `batch_shape x (N + P) x P`
    cav_sum_alpha_log_cdf = torch.sum(cav_alpha_log_cdf, dim=-3).clamp_min(NEG_INF)

    # compute normalization constant Z
    # Equation 35
    cav_log_zeta = torch.log1p(-torch.exp(cav_sum_alpha_log_cdf)).clamp_min(NEG_INF)

    # Need to clamp log values to prevent `exp(-inf) = nan`
    cav_logZ = cav_log_zeta

    # Equation 40 [first bit]
    # `batch_shape x (N + P) x P`
    cav_log_rho = -cav_logZ + cav_sum_alpha_log_cdf

    # Equation 40 [second bit]
    # `batch_shape x M x (N + P) x P`
    cav_log_rho = cav_log_rho.unsqueeze(-3) - cav_alpha_log_cdf + cav_alpha_log_pdf
    cav_rho = -torch.exp(cav_log_rho).clamp(NEG_INF, -NEG_INF)
    ###############################################################################
    # COMPUTE THE PARTIAL DERIVATIVES
    ###############################################################################
    # `batch_shape x M x (N + P) x P x 2`
    # Final vector: `[1, -1]`
    ones_mean = torch.ones(cav_mean_f.shape, **tkwargs)
    ones_mean[..., 1] = -ones_mean[..., 1]

    # `batch_shape x M x (N + P) x P x 2 x 2`
    # Final matrix: `[[1, -1], [-1, 1]]`
    ones_cov = torch.ones(cav_cov_f.shape, **tkwargs)
    ones_cov[..., 0, 1] = -ones_cov[..., 0, 1]
    ones_cov[..., 1, 0] = -ones_cov[..., 1, 0]

    # first partial derivation of the log Z with respect to the mean
    # assuming maximization (this is also where the sign will change)
    # Equation 41
    cav_dlogZ_dm = cav_rho / cav_std_fX_minus_fS
    cav_dlogZ_dm = weight * cav_dlogZ_dm.unsqueeze(-1) * ones_mean

    # second partial derivation of the log Z with respect to the mean
    # Equation 42
    cav_d2logZ_dm2 = -cav_rho * (cav_rho + cav_alpha) / cav_var_fX_minus_fS
    cav_d2logZ_dm2 = cav_d2logZ_dm2.unsqueeze(-1).unsqueeze(-1) * ones_cov

    ###############################################################################
    # COMPUTE THE NEW MEAN AND COVARIANCE
    ###############################################################################
    # compute the new mean and covariance
    cav_updated_mean_f = cav_mean_f + torch.einsum(
        "...ij,...j->...i", cav_cov_f, cav_dlogZ_dm
    )
    cav_updated_cov_f = cav_cov_f + torch.einsum(
        "...ij,...jk,...kl->...il", cav_cov_f, cav_d2logZ_dm2, cav_cov_f
    )
    # transform to natural parameters
    # remove the Pareto diagonal
    cav_updated_cov_f = _replace_pareto_diagonal(cav_updated_cov_f)

    identity = torch.diag_embed(torch.ones(cav_updated_cov_f.shape[:-1], **tkwargs))
    cav_updated_nat_cov_f = torch.inverse(cav_updated_cov_f + jitter * identity)

    cav_updated_nat_mean_f = torch.einsum(
        "...ij,...j->...i", cav_updated_nat_cov_f, cav_updated_mean_f
    )

    # match the moments to compute the gain
    omega_f_nat_mean_new = cav_updated_nat_mean_f - cav_nat_mean_f
    omega_f_nat_cov_new = cav_updated_nat_cov_f - cav_nat_cov_f

    # it is also possible to calculate the update directly as in the original paper:
    # identity = torch.diag_embed(torch.ones(cav_d2logZ_dm2.shape[:-1], **tkwargs))
    # denominator = torch.inverse(cav_cov_f @ cav_d2logZ_dm2 + identity)
    # omega_f_nat_cov_new = - cav_d2logZ_dm2 @ denominator
    # omega_f_nat_mean_new = torch.einsum(
    #     '...ij,...j->...i', denominator,
    #     cav_dlogZ_dm - torch.einsum('...ij,...j->...i', cav_d2logZ_dm2, cav_mean_f)
    # )

    return omega_f_nat_mean_new, omega_f_nat_cov_new


def _safe_update_omega(
    mean_f: Tensor,
    cov_f: Tensor,
    omega_f_nat_mean: Tensor,
    omega_f_nat_cov: Tensor,
    N: int,
    P: int,
    M: int,
    maximize: bool = True,
    jitter: float = 1e-6,
) -> tuple[Tensor, Tensor]:
    r"""Try to update the new omega factors by matching the moments. If the update
    is not possible then this returns the initial omega factors.

    Args:
        mean_f: A `batch_shape x M x (N + P)`-dim Tensor containing the mean vector
            for the objectives.
        cov_f: A `batch_shape x M x (N + P) x (N + P)`-dim Tensor containing the
            covariance matrix for the objectives.
        omega_f_nat_mean: A `batch_shape x M x (N + P) x P x 2`-dim Tensor containing
            the omega natural mean factors for the objective matrix.
        omega_f_nat_cov: A `batch_shape x M x (N + P) x P x 2 x 2`-dim Tensor
            containing the omega natural covariance factors for the objective
            matrix.
        N: The number of design points.
        M: The number of Pareto optimal points.
        M: The number of objectives.
        maximize: If true, we consider the Pareto maximum domination relation.
        jitter: The jitter for the matrix inverse.

    Return:
        A two-element tuple containing

        - omega_f_nat_mean_new: A `batch_shape x M x (N + P) x P x 2` containing the
            new omega natural mean factors for the objective matrix.
        - omega_f_nat_cov_new: A `batch_shape x M x (N + P) x P x 2 x 2` containing
            the new omega natural covariance factors for the objective matrix.
    """
    try:
        omega_f_nat_mean_new, omega_f_nat_cov_new = _update_omega(
            mean_f=mean_f,
            cov_f=cov_f,
            omega_f_nat_mean=omega_f_nat_mean,
            omega_f_nat_cov=omega_f_nat_cov,
            N=N,
            P=P,
            M=M,
            maximize=maximize,
            jitter=jitter,
        )
        check_no_nans(omega_f_nat_mean_new)
        check_no_nans(omega_f_nat_cov_new)
        return omega_f_nat_mean_new, omega_f_nat_cov_new

    except (RuntimeError, InputDataError):
        return omega_f_nat_mean, omega_f_nat_cov


def _update_marginals(
    pred_f_nat_mean: Tensor,
    pred_f_nat_cov: Tensor,
    omega_f_nat_mean: Tensor,
    omega_f_nat_cov: Tensor,
    N: int,
    P: int,
) -> tuple[Tensor, Tensor]:
    r"""Computes the new marginal by summing up all the natural factors.

    Args:
        pred_f_nat_mean: A `batch_shape x M x (N + P)`-dim Tensor containing the
            natural predictive mean matrix for the objectives.
        pred_f_nat_cov: A `batch_shape x M x (N + P) x (N + P)`-dim Tensor containing
            the natural predictive covariance matrix for the objectives.
        omega_f_nat_mean: A `batch_shape x M x (N + P) x P x 2`-dim Tensor containing
            the omega natural mean factors for the objective matrix.
        omega_f_nat_cov: A `batch_shape x M x (N + P) x P x 2 x 2`-dim Tensor
            containing the omega natural covariance factors for the objective matrix.
        N: The number of design points.
        P: The number of Pareto optimal points.

    Returns:
        A two-element tuple containing

        - nat_mean_f: A `batch_shape x M x (N + P)`-dim Tensor containing the updated
            natural mean matrix for the objectives.
        - nat_cov_f: A `batch_shape x M x (N + P) x (N + P)`-dim Tensor containing
            the updated natural predictive covariance matrix for the objectives.
    """

    # `batch_shape x M x (N + P)`
    nat_mean_f = pred_f_nat_mean.clone()
    # `batch_shape x M x (N + P) x (N + P)
    nat_cov_f = pred_f_nat_cov.clone()

    ################################################################################
    # UPDATE THE OBJECTIVES
    ################################################################################
    # remove Pareto diagonal
    # zero out the diagonal
    omega_f_nat_mean[..., range(N, N + P), range(P), :] = 0
    omega_f_nat_cov[..., range(N, N + P), range(P), :, :] = 0

    # `batch_shape x M x (N + P)`
    # sum over the pareto dim
    nat_mean_f = nat_mean_f + omega_f_nat_mean[..., 0].sum(dim=-1)
    # `batch_shape x M x P`
    # sum over the data dim
    nat_mean_f[..., N:] = nat_mean_f[..., N:] + omega_f_nat_mean[..., 1].sum(dim=-2)

    # `batch_shape x M x (N + P)`
    nat_cov_f[..., range(N + P), range(N + P)] = nat_cov_f[
        ..., range(N + P), range(N + P)
    ] + omega_f_nat_cov[..., 0, 0].sum(dim=-1)
    # `batch_shape x M x P`
    nat_cov_f[..., range(N, N + P), range(N, N + P)] = nat_cov_f[
        ..., range(N, N + P), range(N, N + P)
    ] + omega_f_nat_cov[..., 1, 1].sum(dim=-2)

    for p in range(P):
        # `batch_shape x M x (N + P)`
        nat_cov_f[..., range(N + P), N + p] = (
            nat_cov_f[..., range(N + P), N + p] + omega_f_nat_cov[..., p, 0, 1]
        )

        # `batch_shape x M x (N + P)`
        nat_cov_f[..., N + p, range(N + P)] = (
            nat_cov_f[..., N + p, range(N + P)] + omega_f_nat_cov[..., p, 1, 0]
        )

    return nat_mean_f, nat_cov_f


def _damped_update(
    old_factor: Tensor,
    new_factor: Tensor,
    damping_factor: Tensor,
) -> Tensor:
    r"""Computes the damped updated for natural factor.

    Args:
        old_factor: A `batch_shape x param_shape`-dim Tensor containing the old
            natural factor.
        new_factor: A `batch_shape x param_shape`-dim Tensor containing the new
            natural factor.
        damping_factor: A `batch_shape`-dim Tensor containing the damping factor.

    Returns:
        A `batch_shape x param_shape`-dim Tensor containing the updated natural
        factor.
    """
    bs = damping_factor.shape
    fs = old_factor.shape

    df = damping_factor
    for _ in range(len(fs[len(bs) :])):
        df = df.unsqueeze(-1)

    return df * new_factor + (1 - df) * old_factor


def _update_damping(
    nat_cov: Tensor,
    nat_cov_new: Tensor,
    damping_factor: Tensor,
    jitter: Tensor,
) -> tuple[Tensor, Tensor]:
    r"""Updates the damping factor whilst ensuring the covariance matrix is positive
    definite by trying a Cholesky decomposition.

    Args:
        nat_cov: A `batch_shape x R x R`-dim Tensor containing the old natural
            covariance matrix.
        nat_cov_new: A `batch_shape x R x R`-dim Tensor containing the new natural
            covariance matrix.
        damping_factor: A`batch_shape`-dim Tensor containing the damping factor.
        jitter: The amount of jitter added before matrix inversion.

    Returns:
        A two-element tuple containing

        - A `batch_shape x param_shape`-dim Tensor containing the updated damping
            factor.
        - A `batch_shape x R x R`-dim Tensor containing the Cholesky factor.
    """
    tkwargs = {"dtype": nat_cov.dtype, "device": nat_cov.device}
    df = damping_factor
    jitter = jitter * torch.diag_embed(torch.ones(nat_cov.shape[:-1], **tkwargs))
    _, info = torch.linalg.cholesky_ex(nat_cov + jitter)

    if torch.sum(info) > 1:
        raise ValueError(
            "The previous covariance is not positive semi-definite. "
            "This usually happens if the predictive covariance is "
            "ill-conditioned and the added jitter is insufficient."
        )

    damped_nat_cov = _damped_update(
        old_factor=nat_cov, new_factor=nat_cov_new, damping_factor=df
    )
    cholesky_factor, info = torch.linalg.cholesky_ex(damped_nat_cov)
    contains_nans = torch.any(torch.isnan(cholesky_factor)).item()

    run = 0
    while torch.sum(info) > 1 or contains_nans:
        # propose an alternate damping factor which is half the original
        df_alt = 0.5 * df
        # hard threshold at 1e-3
        df_alt = torch.where(
            df_alt > 1e-3, df_alt, torch.zeros(df_alt.shape, **tkwargs)
        )
        # only change the damping factor where psd failure occurs
        df_new = torch.where(info == 0, df, df_alt)

        # new damped covariance
        damped_nat_cov = _damped_update(nat_cov, nat_cov_new, df_new)

        # try cholesky decomposition
        cholesky_factor, info = torch.linalg.cholesky_ex(damped_nat_cov + jitter)
        contains_nans = torch.any(torch.isnan(cholesky_factor)).item()
        df = df_new
        run = run + 1

    return df, cholesky_factor


def _update_damping_when_converged(
    mean_old: Tensor,
    mean_new: Tensor,
    cov_old: Tensor,
    cov_new: Tensor,
    damping_factor: Tensor,
    iteration: Tensor,
    threshold: float = 1e-3,
) -> tuple[Tensor, Tensor, Tensor]:
    r"""Set the damping factor to 0 once converged. Convergence is determined by the
    relative change in the entries of the mean and covariance matrix.

    Args:
        mean_old: A `batch_shape x R`-dim Tensor containing the old natural mean
            matrix for the objective.
        mean_new: A `batch_shape x R`-dim Tensor containing the new natural mean
            matrix for the objective.
        cov_old: A `batch_shape x R x R`-dim Tensor containing the old natural
            covariance matrix for the objective.
        cov_new: A `batch_shape x R x R`-dim Tensor containing the new natural
            covariance matrix for the objective.
        iteration: The current iteration number
        damping_factor: A `batch_shape`-dim Tensor containing the damping factor.

    Returns:
        - A `batch_shape x param_shape`-dim Tensor containing the updated damping
        factor.
        - Difference between `mean_new` and `mean_old`
        - Difference between `cov_new` and `cov_old`
    """
    df = damping_factor.clone()
    delta_mean = mean_new - mean_old
    delta_cov = cov_new - cov_old
    am = torch.amax(abs(mean_old), dim=-1)
    ac = torch.amax(abs(cov_old), dim=(-2, -1))

    if iteration > 2:
        mask_mean = torch.amax(abs(delta_mean), dim=-1) < threshold * am
        mask_cov = torch.amax(abs(delta_cov), dim=(-2, -1)) < threshold * ac
        mask = torch.logical_and(mask_mean, mask_cov)
        df[mask] = 0

    return df, delta_mean, delta_cov


def _augment_factors_with_cached_factors(
    q: int,
    N: int,
    omega_f_nat_mean: Tensor,
    cached_omega_f_nat_mean: Tensor,
    omega_f_nat_cov: Tensor,
    cached_omega_f_nat_cov: Tensor,
) -> tuple[Tensor, Tensor]:
    r"""Incorporate the cached Pareto updated factors in the forward call and
    augment them with the previously computed factors.

    Args:
        q: The batch size.
        N: The number of training points.
        omega_f_nat_mean: A `batch_shape x num_pareto_samples x M x (q + P) x P x 2`
            -dim Tensor containing the omega natural mean for the objective at `X`.
        cached_omega_f_nat_mean: A `num_pareto_samples x M x (N + P) x P x 2`-dim
            Tensor containing the omega natural mean for the objective at `X`.
        omega_f_nat_cov: A `batch_shape x num_pareto_samples x M x (q + P) x P x 2
            x 2` -dim Tensor containing the omega natural covariance for the
            objective at `X`.
        cached_omega_f_nat_cov: A `num_pareto_samples x M x (N + P) x P x 2 x 2`-dim
            Tensor containing the omega covariance mean for the objective at `X`.

    Returns:
        A two-element tuple containing

        - omega_f_nat_mean_new: A `batch_shape x num_pareto_samples x M x (q + P)
            x P x 2`-dim Tensor containing the omega natural mean for the objective
            at `X`.
        - omega_f_nat_cov_new: A `batch_shape x num_pareto_samples x M x (q + P) x
            P x 2 x 2`-dim Tensor containing the omega natural covariance for the
            objective at `X`.
    """
    ##############################################################################
    # omega_f_nat_mean
    ##############################################################################
    # retrieve the natural mean contribution of the Pareto block omega(x_p, x_p) for
    # the objective
    exp_cached_omega_f_nat_mean = cached_omega_f_nat_mean[..., N:, :, :].expand(
        omega_f_nat_mean[..., q:, :, :].shape
    )
    omega_f_nat_mean[..., q:, :, :] = exp_cached_omega_f_nat_mean
    ##############################################################################
    # omega_f_nat_cov
    ##############################################################################
    # retrieve the natural covariance contribution of the Pareto block
    # omega(x_p, x_p) for the objective
    exp_omega_f_nat_cov = cached_omega_f_nat_cov[..., N:, :, :, :].expand(
        omega_f_nat_cov[..., q:, :, :, :].shape
    )
    omega_f_nat_cov[..., q:, :, :, :] = exp_omega_f_nat_cov

    return omega_f_nat_mean, omega_f_nat_cov


def _compute_log_determinant(cov: Tensor, q: int) -> Tensor:
    r"""Computes the sum of the log determinants of a block diagonal covariance
    matrices averaged over the Pareto samples.

    Args:
        cov: A `batch_shape x num_pareto_samples x num_outputs x (q + P) x (q + P)`
            -dim Tensor containing the covariance matrices.
        q: The batch size.

    Return:
        log_det_cov: A `batch_shape`-dim Tensor containing the sum of the
            determinants for each output.
    """
    log_det_cov = torch.logdet(cov[..., 0:q, 0:q])
    check_no_nans(log_det_cov)

    return log_det_cov.sum(dim=-1).mean(dim=-1)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import torch

from botorch.utils.testing import BotorchTestCase
from botorch_community.models.vbll_helper import (
    DenseNormal,
    DenseNormalPrec,
    get_parameterization,
    LowRankNormal,
    Normal,
    tp,
)


def _build_loc_and_cholesky(shape, **tkwargs):
    """Small helper for loc and chol gen."""
    loc = torch.randn(shape, **tkwargs)
    batch_shape = shape[:-1]
    dim = shape[-1]
    L = torch.randn(*batch_shape, dim, dim, **tkwargs)
    L = torch.tril(L)
    diag_idx = torch.arange(dim, device=loc.device)
    L[..., diag_idx, diag_idx] = (
        torch.rand(*batch_shape, dim, **tkwargs) + 0.1
    )  # Ensure positive diagonals
    return loc, L


class VBLLHelperTestCase(BotorchTestCase):
    def setUp(self):
        super().setUp()
        self.tkwargs = {"dtype": torch.float64, "device": self.device}


class TestNormal(VBLLHelperTestCase):
    def test_initialization(self):
        # Test with scalar inputs
        loc = torch.tensor(0.0, **self.tkwargs)
        scale = torch.tensor(1.0, **self.tkwargs)
        dist = Normal(loc, scale)
        self.assertEqual(dist.loc, loc)
        self.assertEqual(dist.scale, scale)

        # Test with vector inputs
        loc = torch.tensor([0.0, 1.0, 2.0], **self.tkwargs)
        scale = torch.tensor([1.0, 2.0, 3.0], **self.tkwargs)
        dist = Normal(loc, scale)
        self.assertTrue(torch.all(dist.loc == loc))
        self.assertTrue(torch.all(dist.scale == scale))

        # Test with batch dimensions
        loc = torch.randn(2, 3, **self.tkwargs)
        scale = torch.rand(2, 3, **self.tkwargs) + 0.1  # Ensure positive
        dist = Normal(loc, scale)
        self.assertTrue(torch.all(dist.loc == loc))
        self.assertTrue(torch.all(dist.scale == scale))

    def test_properties(self):
        loc = torch.tensor([0.0, 1.0, 2.0], **self.tkwargs)
        scale = torch.tensor([1.0, 2.0, 3.0], **self.tkwargs)
        dist = Normal(loc, scale)

        # Test mean property
        self.assertTrue(torch.all(dist.mean == loc))

        # Test var property
        expected_var = scale**2
        self.assertTrue(torch.all(dist.var == expected_var))

        # Test covariance_diagonal property
        self.assertTrue(torch.all(dist.covariance_diagonal == expected_var))

        # Test covariance property
        expected_cov = torch.diag_embed(expected_var)
        self.assertTrue(torch.all(dist.covariance == expected_cov))

        # Test precision property
        expected_prec = torch.diag_embed(1.0 / expected_var)
        self.assertTrue(torch.all(dist.precision == expected_prec))

        # Test logdet_covariance property
        expected_logdet = 2 * torch.log(scale).sum(-1)
        self.assertAllClose(dist.logdet_covariance, expected_logdet)

        # Test logdet_precision property
        expected_logdet_prec = -2 * torch.log(scale).sum(-1)
        self.assertAllClose(dist.logdet_precision, expected_logdet_prec)

        # Test trace_covariance property
        expected_trace = expected_var.sum(-1)
        self.assertAllClose(dist.trace_covariance, expected_trace)

        # Test trace_precision property
        expected_trace_prec = (1.0 / expected_var).sum(-1)
        self.assertAllClose(dist.trace_precision, expected_trace_prec)

        # Test chol_covariance property
        expected_chol = torch.diag_embed(scale)
        self.assertTrue(torch.all(dist.chol_covariance == expected_chol))

    def test_inner_products(self):
        loc = torch.tensor([0.0, 1.0, 2.0], **self.tkwargs)
        scale = torch.tensor([1.0, 2.0, 3.0], **self.tkwargs)
        dist = Normal(loc, scale)

        # Test vector for inner products
        b = torch.tensor([[1.0], [2.0], [3.0]], **self.tkwargs)

        # Test covariance_weighted_inner_prod
        expected_cov_prod = (dist.var.unsqueeze(-1) * (b**2)).sum(-2)
        actual_cov_prod = dist.covariance_weighted_inner_prod(b)
        self.assertAllClose(actual_cov_prod, expected_cov_prod.squeeze(-1))

        # Test precision_weighted_inner_prod
        expected_prec_prod = ((b**2) / dist.var.unsqueeze(-1)).sum(-2)
        actual_prec_prod = dist.precision_weighted_inner_prod(b)
        self.assertAllClose(actual_prec_prod, expected_prec_prod.squeeze(-1))

    def test_addition(self):
        # Test addition with another Normal
        loc1 = torch.tensor([0.0, 1.0, 2.0], **self.tkwargs)
        scale1 = torch.tensor([1.0, 2.0, 3.0], **self.tkwargs)
        dist1 = Normal(loc1, scale1)

        loc2 = torch.tensor([3.0, 2.0, 1.0], **self.tkwargs)
        scale2 = torch.tensor([3.0, 2.0, 1.0], **self.tkwargs)
        dist2 = Normal(loc2, scale2)

        result = dist1 + dist2

        # Check result type
        self.assertIsInstance(result, Normal)

        # Check mean
        expected_mean = loc1 + loc2
        self.assertTrue(torch.all(result.mean == expected_mean))

        # Check scale (variance addition)
        expected_var = dist1.var + dist2.var
        expected_scale = torch.sqrt(torch.clip(expected_var, min=1e-12))
        self.assertAllClose(result.scale, expected_scale)

        # Test addition with a tensor
        tensor = torch.tensor([1.0, 2.0, 3.0], **self.tkwargs)
        result = dist1 + tensor

        # Check result type
        self.assertIsInstance(result, Normal)

        # Check mean
        expected_mean = loc1 + tensor
        self.assertTrue(torch.all(result.mean == expected_mean))

        # Check scale (should be unchanged)
        self.assertTrue(torch.all(result.scale == scale1))

        # Test invalid addition
        with self.assertRaises(NotImplementedError):
            _ = dist1 + 5.0

    def test_matmul(self):
        loc = torch.tensor([0.0, 1.0, 2.0], **self.tkwargs)
        scale = torch.tensor([1.0, 2.0, 3.0], **self.tkwargs)
        dist = Normal(loc, scale)

        # Test matrix multiplication with a vector
        vec = torch.tensor([[2.0], [3.0], [4.0]], **self.tkwargs)
        result = dist @ vec

        # Check result type
        self.assertIsInstance(result, Normal)

        # Check mean
        expected_mean = loc @ vec
        self.assertTrue(torch.all(result.mean == expected_mean))

        # Check scale
        new_cov = dist.covariance_weighted_inner_prod(
            vec.unsqueeze(-3), reduce_dim=False
        )
        expected_scale = torch.sqrt(torch.clip(new_cov, min=1e-12))
        self.assertAllClose(result.scale, expected_scale)

    def test_squeeze(self):
        # Create a distribution with an extra dimension
        loc = torch.tensor([[0.0, 1.0, 2.0]], **self.tkwargs)
        scale = torch.tensor([[1.0, 2.0, 3.0]], **self.tkwargs)
        dist = Normal(loc, scale)

        # Squeeze the first dimension
        result = dist.squeeze(0)

        # Check result type
        self.assertIsInstance(result, Normal)

        # Check dimensions
        self.assertEqual(result.loc.shape, torch.Size([3]))
        self.assertEqual(result.scale.shape, torch.Size([3]))

        # Check values
        self.assertTrue(torch.all(result.loc == loc.squeeze(0)))
        self.assertTrue(torch.all(result.scale == scale.squeeze(0)))


class TestDenseNormal(VBLLHelperTestCase):
    def test_initialization(self):
        # Test with vector input for loc and matrix for cholesky
        loc, L = _build_loc_and_cholesky((3,), **self.tkwargs)
        dist = DenseNormal(loc, L)

        self.assertTrue(torch.all(dist.loc == loc))
        self.assertTrue(torch.all(dist.scale_tril == L))

        # Test with batch dimensions
        loc, L = _build_loc_and_cholesky((2, 3), **self.tkwargs)
        dist = DenseNormal(loc, L)

        self.assertTrue(torch.all(dist.loc == loc))
        self.assertTrue(torch.all(dist.scale_tril == L))

    def test_properties(self):
        loc, L = _build_loc_and_cholesky((3,), **self.tkwargs)
        dist = DenseNormal(loc, L)

        # Test mean property
        self.assertTrue(torch.all(dist.mean == loc))

        # Test chol_covariance property
        self.assertTrue(torch.all(dist.chol_covariance == L))

        # Test covariance property
        expected_cov = L @ tp(L)
        self.assertAllClose(dist.covariance, expected_cov)

        # Test logdet_covariance property
        expected_logdet = 2.0 * torch.diagonal(L, dim1=-2, dim2=-1).log().sum(-1)
        self.assertAllClose(dist.logdet_covariance, expected_logdet)

        # Test trace_covariance property
        expected_trace = (L**2).sum(-1).sum(-1)
        self.assertAllClose(dist.trace_covariance, expected_trace)

    def test_inner_products(self):
        loc, L = _build_loc_and_cholesky((3,), **self.tkwargs)
        dist = DenseNormal(loc, L)

        # Test vector for inner products
        b = torch.tensor([[1.0], [2.0], [3.0]], **self.tkwargs)

        # Test covariance_weighted_inner_prod
        expected_cov_prod = ((tp(L) @ b) ** 2).sum(-2)
        actual_cov_prod = dist.covariance_weighted_inner_prod(b)
        self.assertAllClose(actual_cov_prod, expected_cov_prod.squeeze(-1))

        # Test precision_weighted_inner_prod
        expected_prec_prod = (
            torch.linalg.solve_triangular(L, b, upper=False) ** 2
        ).sum(-2)
        actual_prec_prod = dist.precision_weighted_inner_prod(b)
        self.assertAllClose(actual_prec_prod, expected_prec_prod.squeeze(-1))

    def test_inverse_covariance(self):
        # Test the inverse_covariance property that uses solve_triangular
        loc, L = _build_loc_and_cholesky((3,), **self.tkwargs)
        dist = DenseNormal(loc, L)

        # Compute expected inverse_covariance
        Eye = torch.eye(
            L.shape[-1],
            device=L.device,
            dtype=L.dtype,
        )
        W = torch.linalg.solve_triangular(L, Eye, upper=False)
        expected_inv_cov = tp(W) @ W

        # Compare with the property
        actual_inv_cov = dist.inverse_covariance
        self.assertAllClose(actual_inv_cov, expected_inv_cov)

        # verify that inverse_covariance * covariance ≈ identity
        cov = dist.covariance
        identity_approx = actual_inv_cov @ cov
        eye = torch.eye(3, dtype=torch.float64)
        self.assertAllClose(identity_approx, eye, rtol=1e-5, atol=1e-5)

    def test_matmul(self):
        loc, L = _build_loc_and_cholesky((3,), **self.tkwargs)
        dist = DenseNormal(loc, L)

        # Test matrix multiplication with a vector
        vec = torch.tensor([[2.0], [3.0], [4.0]], **self.tkwargs)
        result = dist @ vec

        # Check result type
        self.assertIsInstance(result, Normal)

        # Check mean
        expected_mean = loc @ vec
        self.assertTrue(torch.all(result.mean == expected_mean))

        # Check scale
        new_cov = dist.covariance_weighted_inner_prod(
            vec.unsqueeze(-3), reduce_dim=False
        )
        expected_scale = torch.sqrt(torch.clip(new_cov, min=1e-12))
        self.assertAllClose(result.scale, expected_scale)

    def test_squeeze(self):
        # Create a distribution with an extra dimension
        loc, L = _build_loc_and_cholesky((1, 3), **self.tkwargs)
        dist = DenseNormal(loc, L)

        # Squeeze the first dimension
        result = dist.squeeze(0)

        # Check result type
        self.assertIsInstance(result, DenseNormal)

        # Check dimensions
        self.assertEqual(result.loc.shape, torch.Size([3]))
        self.assertEqual(result.scale_tril.shape, torch.Size([3, 3]))

        # Check values
        self.assertTrue(torch.all(result.loc == loc.squeeze(0)))
        self.assertTrue(torch.all(result.scale_tril == L.squeeze(0)))


class TestLowRankNormal(VBLLHelperTestCase):
    def test_initialization(self):
        # Test with vector input for loc and matrices for cov_factor and diag
        loc = torch.tensor([0.0, 1.0, 2.0], **self.tkwargs)
        cov_factor = torch.randn(
            3, 2, **self.tkwargs
        )  # 3-dim vector with rank 2 factor
        diag = torch.tensor([1.0, 2.0, 3.0], **self.tkwargs)
        dist = LowRankNormal(loc, cov_factor, diag)

        self.assertTrue(torch.all(dist.loc == loc))
        self.assertTrue(torch.all(dist.cov_factor == cov_factor))
        self.assertTrue(torch.all(dist.cov_diag == diag))

        # Test with batch dimensions
        loc = torch.randn(2, 3, **self.tkwargs)
        cov_factor = torch.randn(2, 3, 2, **self.tkwargs)
        diag = torch.rand(2, 3, **self.tkwargs) + 0.1  # Ensure positive
        dist = LowRankNormal(loc, cov_factor, diag)

        self.assertTrue(torch.all(dist.loc == loc))
        self.assertTrue(torch.all(dist.cov_factor == cov_factor))
        self.assertTrue(torch.all(dist.cov_diag == diag))

    def test_properties(self):
        loc = torch.tensor([0.0, 1.0, 2.0], **self.tkwargs)
        cov_factor = torch.randn(3, 2, **self.tkwargs)
        diag = torch.tensor([1.0, 2.0, 3.0], **self.tkwargs)
        dist = LowRankNormal(loc, cov_factor, diag)

        # Test mean property
        self.assertTrue(torch.all(dist.mean == loc))

        # Test NotImplementedError for unimplemented properties
        with self.assertRaises(NotImplementedError):
            _ = dist.chol_covariance

        with self.assertRaises(NotImplementedError):
            _ = dist.inverse_covariance

        # Test logdet_covariance property
        term1 = torch.log(diag).sum(-1)
        arg1 = tp(cov_factor) @ (cov_factor / diag.unsqueeze(-1))
        term2 = torch.linalg.det(
            arg1 + torch.eye(arg1.shape[-1], dtype=torch.float64)
        ).log()
        expected_logdet = term1 + term2
        self.assertAllClose(dist.logdet_covariance, expected_logdet)

        # Test trace_covariance property
        expected_trace_diag = diag.sum(-1)
        expected_trace_lowrank = (cov_factor**2).sum(-1).sum(-1)
        expected_trace = expected_trace_diag + expected_trace_lowrank
        self.assertAllClose(dist.trace_covariance, expected_trace)

    def test_inner_products(self):
        loc = torch.tensor([0.0, 1.0, 2.0], **self.tkwargs)
        cov_factor = torch.randn(3, 2, **self.tkwargs)
        diag = torch.tensor([1.0, 2.0, 3.0], **self.tkwargs)
        dist = LowRankNormal(loc, cov_factor, diag)

        # Test vector for inner products
        b = torch.tensor([[1.0], [2.0], [3.0]], **self.tkwargs)

        # Test covariance_weighted_inner_prod
        diag_term = (diag.unsqueeze(-1) * (b**2)).sum(-2)
        factor_term = ((tp(cov_factor) @ b) ** 2).sum(-2)
        expected_cov_prod = diag_term + factor_term
        actual_cov_prod = dist.covariance_weighted_inner_prod(b)
        self.assertAllClose(actual_cov_prod, expected_cov_prod.squeeze(-1))

        # Test precision_weighted_inner_prod raises NotImplementedError
        with self.assertRaises(NotImplementedError):
            _ = dist.precision_weighted_inner_prod(b)

    def test_matmul(self):
        loc = torch.tensor([0.0, 1.0, 2.0], **self.tkwargs)
        cov_factor = torch.randn(3, 2, **self.tkwargs)
        diag = torch.tensor([1.0, 2.0, 3.0], **self.tkwargs)
        dist = LowRankNormal(loc, cov_factor, diag)

        # Test matrix multiplication with a vector
        vec = torch.tensor([[2.0], [3.0], [4.0]], **self.tkwargs)
        result = dist @ vec

        # Check result type
        self.assertIsInstance(result, Normal)

        # Check mean
        expected_mean = loc @ vec
        self.assertTrue(torch.all(result.mean == expected_mean))

        # Check scale
        new_cov = dist.covariance_weighted_inner_prod(
            vec.unsqueeze(-3), reduce_dim=False
        )
        expected_scale = torch.sqrt(torch.clip(new_cov, min=1e-12))
        self.assertAllClose(result.scale, expected_scale)

    def test_squeeze(self):
        # Create a distribution with an extra dimension
        loc = torch.tensor([[0.0, 1.0, 2.0]], **self.tkwargs)
        cov_factor = torch.randn(1, 3, 2, **self.tkwargs)
        diag = torch.tensor([[1.0, 2.0, 3.0]], **self.tkwargs)
        dist = LowRankNormal(loc, cov_factor, diag)

        # Squeeze the first dimension
        result = dist.squeeze(0)

        # Check result type
        self.assertIsInstance(result, LowRankNormal)

        # Check dimensions
        self.assertEqual(result.loc.shape, torch.Size([3]))
        self.assertEqual(result.cov_factor.shape, torch.Size([3, 2]))
        self.assertEqual(result.cov_diag.shape, torch.Size([3]))

        # Check values
        self.assertTrue(torch.all(result.loc == loc.squeeze(0)))
        self.assertTrue(torch.all(result.cov_factor == cov_factor.squeeze(0)))
        self.assertTrue(torch.all(result.cov_diag == diag.squeeze(0)))


class TestDenseNormalPrec(VBLLHelperTestCase):
    def test_initialization(self):
        # Test with vector input for loc and matrix for cholesky
        loc, L = _build_loc_and_cholesky((3,), **self.tkwargs)
        dist = DenseNormalPrec(loc, L)

        self.assertTrue(torch.all(dist.loc == loc))
        self.assertTrue(torch.all(dist.tril == L))

        # Test with batch dimensions
        loc, L = _build_loc_and_cholesky((2, 3), **self.tkwargs)
        dist = DenseNormalPrec(loc, L)

        self.assertTrue(torch.all(dist.loc == loc))
        self.assertTrue(torch.all(dist.tril == L))

    def test_properties(self):
        loc, L = _build_loc_and_cholesky((3,), **self.tkwargs)
        dist = DenseNormalPrec(loc, L)

        # Test mean property
        self.assertTrue(torch.all(dist.mean == loc))

        # Test inverse_covariance property
        expected_prec = L @ tp(L)
        self.assertAllClose(dist.inverse_covariance, expected_prec)

        # Test NotImplementedError for chol_covariance
        with self.assertRaises(NotImplementedError):
            _ = dist.chol_covariance

        # Test logdet_covariance property
        expected_logdet = -2.0 * torch.diagonal(L, dim1=-2, dim2=-1).log().sum(-1)
        self.assertAllClose(dist.logdet_covariance, expected_logdet)

    def test_inner_products(self):
        loc, L = _build_loc_and_cholesky((3,), **self.tkwargs)
        dist = DenseNormalPrec(loc, L)

        # Test vector for inner products
        b = torch.tensor([[1.0], [2.0], [3.0]], **self.tkwargs)

        # Test covariance_weighted_inner_prod
        expected_cov_prod = (torch.linalg.solve(L, b) ** 2).sum(-2)
        actual_cov_prod = dist.covariance_weighted_inner_prod(b)
        self.assertAllClose(actual_cov_prod, expected_cov_prod.squeeze(-1))

        # Test precision_weighted_inner_prod
        expected_prec_prod = ((tp(L) @ b) ** 2).sum(-2)
        actual_prec_prod = dist.precision_weighted_inner_prod(b)
        self.assertAllClose(actual_prec_prod, expected_prec_prod.squeeze(-1))

    def test_covariance(self):
        # Test the covariance property that uses cholesky_inverse
        loc, L = _build_loc_and_cholesky((3,), **self.tkwargs)
        dist = DenseNormalPrec(loc, L)

        # Compute expected covariance using cholesky_inverse
        expected_cov = torch.cholesky_inverse(L)

        # Compare with the property
        actual_cov = dist.covariance
        self.assertAllClose(actual_cov, expected_cov)

        # verify that covariance * precision_matrix ≈ identity
        precision = dist.inverse_covariance
        identity_approx = actual_cov @ precision
        eye = torch.eye(3, dtype=torch.float64)
        self.assertAllClose(identity_approx, eye, rtol=1e-5, atol=1e-5)

        # check that the covariance is symmetric
        self.assertAllClose(actual_cov, actual_cov.t())

    def test_matmul(self):
        loc, L = _build_loc_and_cholesky((3,), **self.tkwargs)
        dist = DenseNormalPrec(loc, L)

        # Test matrix multiplication with a vector
        vec = torch.tensor([[2.0], [3.0], [4.0]], **self.tkwargs)
        result = dist @ vec

        # Check result type
        self.assertIsInstance(result, Normal)

        # Check mean
        expected_mean = loc @ vec
        self.assertTrue(torch.all(result.mean == expected_mean))

        # Check scale
        new_cov = dist.covariance_weighted_inner_prod(
            vec.unsqueeze(-3), reduce_dim=False
        )
        expected_scale = torch.sqrt(torch.clip(new_cov, min=1e-12))
        self.assertAllClose(result.scale, expected_scale)

    def test_squeeze(self):
        # Create a distribution with an extra dimension
        loc, L = _build_loc_and_cholesky((1, 3), **self.tkwargs)
        dist = DenseNormalPrec(loc, L)

        # Squeeze the first dimension
        result = dist.squeeze(0)

        # Check result type
        self.assertIsInstance(result, DenseNormalPrec)

        # Check dimensions
        self.assertEqual(result.loc.shape, torch.Size([3]))
        self.assertEqual(result.tril.shape, torch.Size([3, 3]))

        # Check values
        self.assertTrue(torch.all(result.loc == loc.squeeze(0)))
        self.assertTrue(torch.all(result.tril == L.squeeze(0)))


class TestGetParameterization(BotorchTestCase):
    def test_get_parameterization(self):
        # Test all valid parameterizations
        self.assertEqual(get_parameterization("dense"), DenseNormal)
        self.assertEqual(get_parameterization("dense_precision"), DenseNormalPrec)
        self.assertEqual(get_parameterization("diagonal"), Normal)
        self.assertEqual(get_parameterization("lowrank"), LowRankNormal)

        # Test invalid parameterization
        with self.assertRaises(ValueError):
            get_parameterization("invalid_param")



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import torch
from botorch.utils.testing import BotorchTestCase
from botorch_community.models.example import ExampleModel
from gpytorch.kernels import RBFKernel, ScaleKernel


class TestExampleModel(BotorchTestCase):
    def test_example_gp(self) -> None:
        model = ExampleModel(
            train_X=torch.rand(2, 2),
            train_Y=torch.rand(2, 1),
        )
        self.assertIsInstance(model.covar_module, ScaleKernel)
        self.assertIsInstance(model.covar_module.base_kernel, RBFKernel)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Callable, Iterable, Iterator, Mapping
from typing import Any

from botorch.exceptions.errors import UnsupportedError
from botorch.sampling.pathwise.features import FeatureMap
from botorch.sampling.pathwise.utils import (
    TInputTransform,
    TOutputTransform,
    TransformedModuleMixin,
)
from torch import Tensor
from torch.nn import Module, ModuleDict, ModuleList, Parameter


class SamplePath(ABC, TransformedModuleMixin, Module):
    r"""Abstract base class for Botorch sample paths."""

    @abstractmethod
    def set_ensemble_as_batch(self, ensemble_as_batch: bool) -> None:
        """Sets whether the ensemble dimension is considered as a batch dimension.

        Args:
            ensemble_as_batch: Whether the ensemble dimension is considered as a batch
                dimension or not.
        """
        pass  # pragma: no cover


class PathDict(SamplePath):
    r"""A dictionary of SamplePaths."""

    def __init__(
        self,
        paths: Mapping[str, SamplePath] | None = None,
        join: Callable[[list[Tensor]], Tensor] | None = None,
        input_transform: TInputTransform | None = None,
        output_transform: TOutputTransform | None = None,
    ) -> None:
        r"""Initializes a PathDict instance.

        Args:
            paths: An optional mapping of strings to sample paths.
            join: An optional callable used to combine each path's outputs.
            input_transform: An optional input transform for the module.
            output_transform: An optional output transform for the module.
        """
        if join is None and output_transform is not None:
            raise UnsupportedError("Output transforms must be preceded by a join rule.")

        super().__init__()
        self.join = join
        self.input_transform = input_transform
        self.output_transform = output_transform
        self.paths = (
            paths
            if isinstance(paths, ModuleDict)
            else ModuleDict({} if paths is None else paths)
        )

    def forward(self, x: Tensor, **kwargs: Any) -> Tensor | dict[str, Tensor]:
        out = [path(x, **kwargs) for path in self.paths.values()]
        return dict(zip(self.paths, out)) if self.join is None else self.join(out)

    def items(self) -> Iterable[tuple[str, SamplePath]]:
        return self.paths.items()

    def keys(self) -> Iterable[str]:
        return self.paths.keys()

    def values(self) -> Iterable[SamplePath]:
        return self.paths.values()

    def __len__(self) -> int:
        return len(self.paths)

    def __iter__(self) -> Iterator[SamplePath]:
        yield from self.paths

    def __delitem__(self, key: str) -> None:
        del self.paths[key]

    def __getitem__(self, key: str) -> SamplePath:
        return self.paths[key]

    def __setitem__(self, key: str, val: SamplePath) -> None:
        self.paths[key] = val

    def set_ensemble_as_batch(self, ensemble_as_batch: bool) -> None:
        """Sets whether the ensemble dimension is considered as a batch dimension.

        Args:
            ensemble_as_batch: Whether the ensemble dimension is considered as a batch
                dimension or not.
        """
        for path in self.paths.values():
            path.set_ensemble_as_batch(ensemble_as_batch)


class PathList(SamplePath):
    r"""A list of SamplePaths."""

    def __init__(
        self,
        paths: Iterable[SamplePath] | None = None,
        join: Callable[[list[Tensor]], Tensor] | None = None,
        input_transform: TInputTransform | None = None,
        output_transform: TOutputTransform | None = None,
    ) -> None:
        r"""Initializes a PathList instance.

        Args:
            paths: An optional iterable of sample paths.
            join: An optional callable used to combine each path's outputs.
            input_transform: An optional input transform for the module.
            output_transform: An optional output transform for the module.
        """

        if join is None and output_transform is not None:
            raise UnsupportedError("Output transforms must be preceded by a join rule.")

        super().__init__()
        self.join = join
        self.input_transform = input_transform
        self.output_transform = output_transform
        self.paths = (
            paths
            if isinstance(paths, ModuleList)
            else ModuleList({} if paths is None else paths)
        )

    def forward(self, x: Tensor, **kwargs: Any) -> Tensor | list[Tensor]:
        out = [path(x, **kwargs) for path in self.paths]
        return out if self.join is None else self.join(out)

    def __len__(self) -> int:
        return len(self.paths)

    def __iter__(self) -> Iterator[SamplePath]:
        yield from self.paths

    def __delitem__(self, key: int) -> None:
        del self.paths[key]

    def __getitem__(self, key: int) -> SamplePath:
        return self.paths[key]

    def __setitem__(self, key: int, val: SamplePath) -> None:
        self.paths[key] = val

    def set_ensemble_as_batch(self, ensemble_as_batch: bool) -> None:
        """Sets whether the ensemble dimension is considered as a batch dimension.

        Args:
            ensemble_as_batch: Whether the ensemble dimension is considered as a batch
                dimension or not.
        """
        for path in self.paths:
            path.set_ensemble_as_batch(ensemble_as_batch)


class GeneralizedLinearPath(SamplePath):
    r"""A sample path in the form of a generalized linear model."""

    def __init__(
        self,
        feature_map: FeatureMap,
        weight: Parameter | Tensor,
        bias_module: Module | None = None,
        input_transform: TInputTransform | None = None,
        output_transform: TOutputTransform | None = None,
        is_ensemble: bool = False,
        ensemble_as_batch: bool = False,
    ):
        r"""Initializes a GeneralizedLinearPath instance.

        .. code-block:: text

            path(x) = output_transform(bias_module(z) + feature_map(z)^T weight),
            where z = input_transform(x).

        Args:
            feature_map: A map used to featurize the module's inputs.
            weight: A tensor of weights used to combine input features. When generated
                with `draw_kernel_feature_paths`, `weight` is a Tensor with the shape
                `sample_shape x batch_shape x num_outputs`.
            bias_module: An optional module used to define additive offsets.
            input_transform: An optional input transform for the module.
            output_transform: An optional output transform for the module.
            is_ensemble: Whether the associated model is an ensemble model or not.
            ensemble_as_batch: Whether the ensemble dimension is added as a batch
                dimension or not. If `True`, the ensemble dimension is treated as a
                batch dimension, which allows for the joint optimization of all members
                of the ensemble.
        """
        super().__init__()
        self.feature_map = feature_map
        if not isinstance(weight, Parameter):
            self.register_buffer("weight", weight)
        self.weight = weight
        self.bias_module = bias_module
        self.input_transform = input_transform
        self.output_transform = output_transform
        self.is_ensemble = is_ensemble
        self.ensemble_as_batch = ensemble_as_batch

    def forward(self, x: Tensor, **kwargs) -> Tensor:
        """Evaluates the path.

        Args:
            x: The input tensor of shape `batch_shape x [num_ensemble x] q x d`, where
                `num_ensemble` is the number of ensemble members and is required to
                *only* be included if `is_ensemble=True` and `ensemble_as_batch=True`.
            kwargs: Additional keyword arguments passed to the feature map.

        Returns:
            A tensor of shape `batch_shape x [num_ensemble x] q x m`, where `m` is the
            number of outputs, where `num_ensemble` is only included if `is_ensemble`
            is `True`, and regardless of whether `ensemble_as_batch` is `True` or not.
        """
        if self.is_ensemble and not self.ensemble_as_batch:
            # assuming that the ensembling dimension is added after (n, d), but
            # before the other batch dimensions, starting from the left.
            x = x.unsqueeze(-3)
        feat = self.feature_map(x, **kwargs)
        out = (feat @ self.weight.unsqueeze(-1)).squeeze(-1)
        return out if self.bias_module is None else out + self.bias_module(x)

    def set_ensemble_as_batch(self, ensemble_as_batch: bool) -> None:
        """Sets whether the ensemble dimension is considered as a batch dimension.

        Args:
            ensemble_as_batch: Whether the ensemble dimension is considered as a batch
                dimension or not.
        """
        self.ensemble_as_batch = ensemble_as_batch



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

r"""
Analytic Acquisition Functions that evaluate the posterior without performing
Monte-Carlo sampling.
"""

from __future__ import annotations

import math

from abc import ABC, abstractmethod
from contextlib import nullcontext
from copy import deepcopy

import torch
from botorch.acquisition.acquisition import AcquisitionFunction
from botorch.acquisition.objective import PosteriorTransform
from botorch.exceptions import UnsupportedError
from botorch.exceptions.warnings import legacy_ei_numerics_warning
from botorch.models.gp_regression import SingleTaskGP
from botorch.models.gpytorch import GPyTorchModel
from botorch.models.model import Model
from botorch.utils.constants import get_constants_like
from botorch.utils.probability import MVNXPB
from botorch.utils.probability.utils import (
    compute_log_prob_feas_from_bounds,
    log_ndtr as log_Phi,
    log_phi,
    ndtr as Phi,
    phi,
)
from botorch.utils.safe_math import log1mexp, logmeanexp
from botorch.utils.transforms import (
    average_over_ensemble_models,
    t_batch_mode_transform,
)
from gpytorch.likelihoods.gaussian_likelihood import FixedNoiseGaussianLikelihood
from torch import Tensor
from torch.nn.functional import pad

# the following two numbers are needed for _log_ei_helper
_neg_inv_sqrt2 = -(2**-0.5)
_log_sqrt_pi_div_2 = math.log(math.pi / 2) / 2


class AnalyticAcquisitionFunction(AcquisitionFunction, ABC):
    """Base class for analytic acquisition functions."""

    def __init__(
        self,
        model: Model,
        posterior_transform: PosteriorTransform | None = None,
        allow_multi_output: bool = False,
    ) -> None:
        r"""Base constructor for analytic acquisition functions.

        Args:
            model: A fitted single-outcome model.
            posterior_transform: A PosteriorTransform. If using a multi-output model,
                a PosteriorTransform that transforms the multi-output posterior into a
                single-output posterior is required.
              allow_multi_output: If False, requires a posterior_transform if a
                multi-output model is passed.
        """
        super().__init__(model=model)
        if posterior_transform is None:
            if not allow_multi_output and model.num_outputs != 1:
                raise UnsupportedError(
                    "Must specify a posterior transform when using a "
                    "multi-output model."
                )
        else:
            if not isinstance(posterior_transform, PosteriorTransform):
                raise UnsupportedError(
                    "AnalyticAcquisitionFunctions only support PosteriorTransforms."
                )
        self.posterior_transform = posterior_transform

    def set_X_pending(self, X_pending: Tensor | None = None) -> None:
        raise UnsupportedError(
            "Analytic acquisition functions do not account for X_pending yet."
        )

    def _mean_and_sigma(
        self, X: Tensor, compute_sigma: bool = True, min_var: float = 1e-12
    ) -> tuple[Tensor, Tensor | None]:
        """Computes the first and second moments of the model posterior.

        Args:
            X: A `(b1 x ... bk) x 1 x d`-dim batched tensor of `d`-dim design points.
            compute_sigma: Boolean indicating whether or not to compute the second
                moment (default: True).
            min_var: The minimum value the variance is clamped too. Should be positive.

        Returns:
            A tuple of tensors of shape `(b1 x ... x bk) x m` containing the first and
            second moments of the model posterior, where `m` is the number of outputs.
            Returns `None` instead of the second tensor if `compute_sigma` is False.
        """
        self.to(X)  # ensures buffers / parameters are on the same device and dtype
        posterior = self.model.posterior(
            X=X, posterior_transform=self.posterior_transform
        )
        mean = posterior.mean.squeeze(-2)  # remove q-batch dimension
        if not compute_sigma:
            return mean, None
        sigma = posterior.variance.clamp_min(min_var).sqrt().view(mean.shape)
        return mean, sigma


class LogProbabilityOfImprovement(AnalyticAcquisitionFunction):
    r"""Single-outcome Log Probability of Improvement.

    Logarithm of the probability of improvement over the current best observed value,
    computed using the analytic formula under a Normal posterior distribution. Only
    supports the case of q=1. Requires the posterior to be Gaussian. The model must be
    single-outcome.

    The logarithm of the probability of improvement is numerically better behaved
    than the original function, which can lead to significantly improved optimization
    of the acquisition function. This is analogous to the common practice of optimizing
    the *log* likelihood of a probabilistic model - rather the likelihood - for the
    sake of maximium likelihood estimation.

    `logPI(x) = log(P(y >= best_f)), y ~ f(x)`

    Example:
        >>> model = SingleTaskGP(train_X, train_Y)
        >>> LogPI = LogProbabilityOfImprovement(model, best_f=0.2)
        >>> log_pi = LogPI(test_X)
    """

    _log: bool = True

    def __init__(
        self,
        model: Model,
        best_f: float | Tensor,
        posterior_transform: PosteriorTransform | None = None,
        maximize: bool = True,
    ):
        r"""Single-outcome Probability of Improvement.

        Args:
            model: A fitted single-outcome model.
            best_f: Either a scalar or a `b`-dim Tensor (batch mode) representing
                the best function value observed so far (assumed noiseless).
            posterior_transform: A PosteriorTransform. If using a multi-output model,
                a PosteriorTransform that transforms the multi-output posterior into a
                single-output posterior is required.
            maximize: If True, consider the problem a maximization problem.
        """
        super().__init__(model=model, posterior_transform=posterior_transform)
        self.register_buffer("best_f", torch.as_tensor(best_f))
        self.maximize = maximize

    @t_batch_mode_transform(expected_q=1)
    @average_over_ensemble_models
    def forward(self, X: Tensor) -> Tensor:
        r"""Evaluate the Log Probability of Improvement on the candidate set X.

        Args:
            X: A `(b1 x ... bk) x 1 x d`-dim batched tensor of `d`-dim design points.

        Returns:
            A `(b1 x ... bk)`-dim tensor of Log Probability of Improvement values at
            the given design points `X`.
        """
        mean, sigma = self._mean_and_sigma(X)  # `(b1 x ... bk) x 1`
        u = _scaled_improvement(mean, sigma, self.best_f, self.maximize)
        return log_Phi(u.squeeze(-1))


class ProbabilityOfImprovement(AnalyticAcquisitionFunction):
    r"""Single-outcome Probability of Improvement.

    Probability of improvement over the current best observed value, computed
    using the analytic formula under a Normal posterior distribution. Only
    supports the case of q=1. Requires the posterior to be Gaussian. The model
    must be single-outcome.

    `PI(x) = P(y >= best_f), y ~ f(x)`

    Example:
        >>> model = SingleTaskGP(train_X, train_Y)
        >>> PI = ProbabilityOfImprovement(model, best_f=0.2)
        >>> pi = PI(test_X)
    """

    def __init__(
        self,
        model: Model,
        best_f: float | Tensor,
        posterior_transform: PosteriorTransform | None = None,
        maximize: bool = True,
    ):
        r"""Single-outcome Probability of Improvement.

        Args:
            model: A fitted single-outcome model.
            best_f: Either a scalar or a `b`-dim Tensor (batch mode) representing
                the best function value observed so far (assumed noiseless).
            posterior_transform: A PosteriorTransform. If using a multi-output model,
                a PosteriorTransform that transforms the multi-output posterior into a
                single-output posterior is required.
            maximize: If True, consider the problem a maximization problem.
        """
        super().__init__(model=model, posterior_transform=posterior_transform)
        self.register_buffer("best_f", torch.as_tensor(best_f))
        self.maximize = maximize

    @t_batch_mode_transform(expected_q=1)
    @average_over_ensemble_models
    def forward(self, X: Tensor) -> Tensor:
        r"""Evaluate the Probability of Improvement on the candidate set X.

        Args:
            X: A `(b1 x ... bk) x 1 x d`-dim batched tensor of `d`-dim design points.

        Returns:
            A `(b1 x ... bk)`-dim tensor of Probability of Improvement values at the
            given design points `X`.
        """
        mean, sigma = self._mean_and_sigma(X)  # `(b1 x ... bk) x 1`
        u = _scaled_improvement(mean, sigma, self.best_f, self.maximize)
        return Phi(u.squeeze(-1))


class qAnalyticProbabilityOfImprovement(AnalyticAcquisitionFunction):
    r"""Approximate, single-outcome batch Probability of Improvement using MVNXPB.

    This implementation uses MVNXPB, a bivariate conditioning algorithm for
    approximating P(a <= Y <= b) for multivariate normal Y.
    See [Trinh2015bivariate]_. This (analytic) approximate q-PI is given by
    `approx-qPI(X) = P(max Y >= best_f) = 1 - P(Y < best_f), Y ~ f(X),
    X = (x_1,...,x_q)`, where `P(Y < best_f)` is estimated using MVNXPB.
    """

    def __init__(
        self,
        model: Model,
        best_f: float | Tensor,
        posterior_transform: PosteriorTransform | None = None,
        maximize: bool = True,
    ) -> None:
        """qPI using an analytic approximation.

        Args:
            model: A fitted single-outcome model.
            best_f: Either a scalar or a `b`-dim Tensor (batch mode) representing
                the best function value observed so far (assumed noiseless).
            posterior_transform: A PosteriorTransform. If using a multi-output model,
                a PosteriorTransform that transforms the multi-output posterior into a
                single-output posterior is required.
            maximize: If True, consider the problem a maximization problem.
        """
        super().__init__(model=model, posterior_transform=posterior_transform)
        self.maximize = maximize
        if not torch.is_tensor(best_f):
            best_f = torch.tensor(best_f)
        self.register_buffer("best_f", best_f)

    @t_batch_mode_transform()
    @average_over_ensemble_models
    def forward(self, X: Tensor) -> Tensor:
        """Evaluate approximate qPI on the candidate set X.

        Args:
            X: A `batch_shape x q x d`-dim Tensor of t-batches with `q` `d`-dim design
                points each

        Returns:
            A `batch_shape`-dim Tensor of approximate Probability of Improvement values
            at the given design points `X`, where `batch_shape'` is the broadcasted
            batch shape of model and input `X`.
        """
        self.best_f = self.best_f.to(X)
        posterior = self.model.posterior(
            X=X, posterior_transform=self.posterior_transform
        )

        covariance = posterior.distribution.covariance_matrix
        bounds = pad(
            (self.best_f.unsqueeze(-1) - posterior.distribution.mean).unsqueeze(-1),
            pad=(1, 0) if self.maximize else (0, 1),
            value=-float("inf") if self.maximize else float("inf"),
        )
        # 1 - P(no improvement over best_f)
        solver = MVNXPB(covariance_matrix=covariance, bounds=bounds)
        return -solver.solve().expm1()


class ExpectedImprovement(AnalyticAcquisitionFunction):
    r"""Single-outcome Expected Improvement (analytic).

    Computes classic Expected Improvement over the current best observed value,
    using the analytic formula for a Normal posterior distribution. Unlike the
    MC-based acquisition functions, this relies on the posterior at single test
    point being Gaussian (and require the posterior to implement `mean` and
    `variance` properties). Only supports the case of `q=1`. The model must be
    single-outcome.

    `EI(x) = E(max(f(x) - best_f, 0)),`

    where the expectation is taken over the value of stochastic function `f` at `x`.

    Example:
        >>> model = SingleTaskGP(train_X, train_Y)
        >>> EI = ExpectedImprovement(model, best_f=0.2)
        >>> ei = EI(test_X)

    NOTE: It is strongly recommended to use LogExpectedImprovement instead of regular
    EI, as it can lead to substantially improved BO performance through improved
    numerics. See https://arxiv.org/abs/2310.20708 for details.
    """

    def __init__(
        self,
        model: Model,
        best_f: float | Tensor,
        posterior_transform: PosteriorTransform | None = None,
        maximize: bool = True,
    ):
        r"""Single-outcome Expected Improvement (analytic).

        Args:
            model: A fitted single-outcome model.
            best_f: Either a scalar or a `b`-dim Tensor (batch mode) representing
                the best function value observed so far (assumed noiseless).
            posterior_transform: A PosteriorTransform. If using a multi-output model,
                a PosteriorTransform that transforms the multi-output posterior into a
                single-output posterior is required.
            maximize: If True, consider the problem a maximization problem.
        """
        legacy_ei_numerics_warning(legacy_name=type(self).__name__)
        super().__init__(model=model, posterior_transform=posterior_transform)
        self.register_buffer("best_f", torch.as_tensor(best_f))
        self.maximize = maximize

    @t_batch_mode_transform(expected_q=1)
    @average_over_ensemble_models
    def forward(self, X: Tensor) -> Tensor:
        r"""Evaluate Expected Improvement on the candidate set X.

        Args:
            X: A `(b1 x ... bk) x 1 x d`-dim batched tensor of `d`-dim design points.
                Expected Improvement is computed for each point individually,
                i.e., what is considered are the marginal posteriors, not the
                joint.

        Returns:
            A `(b1 x ... bk)`-dim tensor of Expected Improvement values at the
            given design points `X`.
        """
        mean, sigma = self._mean_and_sigma(X)  # `(b1 x ... bk) x 1`
        u = _scaled_improvement(mean, sigma, self.best_f, self.maximize)
        return (sigma * _ei_helper(u)).squeeze(-1)


class LogExpectedImprovement(AnalyticAcquisitionFunction):
    r"""Single-outcome Log Expected Improvement (analytic).

    Computes the logarithm of the classic Expected Improvement acquisition function, in
    a numerically robust manner. In particular, the implementation takes special care
    to avoid numerical issues in the computation of the acquisition value and its
    gradient in regions where improvement is predicted to be virtually impossible.

    See [Ament2023logei]_ for details. Formally,

    `LogEI(x) = log(E(max(f(x) - best_f, 0))),`

    where the expectation is taken over the value of stochastic function `f` at `x`.

    Example:
        >>> model = SingleTaskGP(train_X, train_Y)
        >>> LogEI = LogExpectedImprovement(model, best_f=0.2)
        >>> ei = LogEI(test_X)
    """

    _log: bool = True

    def __init__(
        self,
        model: Model,
        best_f: float | Tensor,
        posterior_transform: PosteriorTransform | None = None,
        maximize: bool = True,
    ):
        r"""Logarithm of single-outcome Expected Improvement (analytic).

        Args:
            model: A fitted single-outcome model.
            best_f: Either a scalar or a `b`-dim Tensor (batch mode) representing
                the best function value observed so far (assumed noiseless).
            posterior_transform: A PosteriorTransform. If using a multi-output model,
                a PosteriorTransform that transforms the multi-output posterior into a
                single-output posterior is required.
            maximize: If True, consider the problem a maximization problem.
        """
        super().__init__(model=model, posterior_transform=posterior_transform)
        self.register_buffer("best_f", torch.as_tensor(best_f))
        self.maximize = maximize

    @t_batch_mode_transform(expected_q=1)
    @average_over_ensemble_models
    def forward(self, X: Tensor) -> Tensor:
        r"""Evaluate logarithm of Expected Improvement on the candidate set X.

        Args:
            X: A `(b1 x ... bk) x 1 x d`-dim batched tensor of `d`-dim design points.
                Expected Improvement is computed for each point individually,
                i.e., what is considered are the marginal posteriors, not the
                joint.

        Returns:
            A `(b1 x ... bk)`-dim tensor of the logarithm of the Expected Improvement
            values at the given design points `X`.
        """
        mean, sigma = self._mean_and_sigma(X)  # `(b1 x ... bk) x 1`
        u = _scaled_improvement(mean, sigma, self.best_f, self.maximize)
        return (_log_ei_helper(u) + sigma.log()).squeeze(-1)


class ConstrainedAnalyticAcquisitionFunctionMixin(ABC):
    r"""Base class for constrained analytic acquisition functions."""

    def __init__(
        self,
        constraints: dict[int, tuple[float | None, float | None]],
    ) -> None:
        r"""Analytic Log Probability of Feasibility.

        Args:
            model: A fitted single- or multi-output model.
            constraints: A dictionary of the form `{i: [lower, upper]}`, where
                `i` is the output index, and `lower` and `upper` are lower and upper
                bounds on that output (resp. interpreted as -Inf / Inf if None).
        """
        self.constraints = constraints
        self._preprocess_constraint_bounds(constraints=constraints)

    @abstractmethod
    def register_buffer(self, name: str, value: Tensor) -> None:
        """Add a buffer that can be accessed by `self.name` and stores the Tensor
        `value`, usually provided by derivatives that also inherit from `nn.Module`.
        """

    def _preprocess_constraint_bounds(
        self,
        constraints: dict[int, tuple[float | None, float | None]],
    ) -> None:
        r"""Set up constraint bounds.

        Args:
            constraints: A dictionary of the form `{i: [lower, upper]}`, where
                `i` is the output index, and `lower` and `upper` are lower and upper
                bounds on that output (resp. interpreted as -Inf / Inf if None)
        """
        con_lower, con_lower_inds = [], []
        con_upper, con_upper_inds = [], []
        con_both, con_both_inds = [], []
        con_indices = list(constraints.keys())
        if len(con_indices) == 0:
            raise ValueError("There must be at least one constraint.")
        # CEI, LogCEI have an objective index, but LogPOF does not.
        if hasattr(self, "objective_index") and self.objective_index in con_indices:
            raise ValueError(
                "Output corresponding to objective should not be a constraint."
            )
        for k in con_indices:
            if constraints[k][0] is not None and constraints[k][1] is not None:
                if constraints[k][1] <= constraints[k][0]:
                    raise ValueError("Upper bound is less than the lower bound.")
                con_both_inds.append(k)
                con_both.append([constraints[k][0], constraints[k][1]])
            elif constraints[k][0] is not None:
                con_lower_inds.append(k)
                con_lower.append(constraints[k][0])
            elif constraints[k][1] is not None:
                con_upper_inds.append(k)
                con_upper.append(constraints[k][1])

        for name, value in [
            ("con_lower_inds", con_lower_inds),
            ("con_upper_inds", con_upper_inds),
            ("con_both_inds", con_both_inds),
            ("con_both", con_both),
            ("con_lower", con_lower),
            ("con_upper", con_upper),
        ]:
            # tensor-based indexing is much faster than list-based advanced indexing
            self.register_buffer(name, tensor=torch.as_tensor(value))

    def _compute_log_prob_feas(
        self,
        means: Tensor,
        sigmas: Tensor,
    ) -> Tensor:
        r"""Compute logarithm of the feasibility probability for each batch of X.

        Args:
            means: A `(b) x m`-dim Tensor of means.
            sigmas: A `(b) x m`-dim Tensor of standard deviations.

        Returns:
            A `(b)`-dim tensor of log feasibility probabilities

        Note: This function does case-work for upper bound, lower bound, and both-sided
        bounds. Another way to do it would be to use 'inf' and -'inf' for the
        one-sided bounds and use the logic for the both-sided case. But this
        causes an issue with autograd since we get 0 * inf.
        """
        return compute_log_prob_feas_from_bounds(
            con_lower_inds=self.con_lower_inds,
            con_upper_inds=self.con_upper_inds,
            con_both_inds=self.con_both_inds,
            con_lower=self.con_lower,
            con_upper=self.con_upper,
            con_both=self.con_both,
            means=means,
            sigmas=sigmas,
        )


class LogConstrainedExpectedImprovement(
    AnalyticAcquisitionFunction, ConstrainedAnalyticAcquisitionFunctionMixin
):
    r"""Log Constrained Expected Improvement (feasibility-weighted).

    Computes the logarithm of the analytic expected improvement for a Normal posterior
    distribution weighted by a probability of feasibility. The objective and
    constraints are assumed to be independent and have Gaussian posterior
    distributions. Only supports non-batch mode (i.e. `q=1`). The model should be
    multi-outcome, with the index of the objective and constraints passed to
    the constructor.

    See [Ament2023logei]_ for details. Formally,

    `LogConstrainedEI(x) = log(EI(x)) + Sum_i log(P(y_i \in [lower_i, upper_i]))`,

    where `y_i ~ constraint_i(x)` and `lower_i`, `upper_i` are the lower and
    upper bounds for the i-th constraint, respectively.

    Example:
        # example where the 0th output has a non-negativity constraint and
        # the 1st output is the objective
        >>> model = SingleTaskGP(train_X, train_Y)
        >>> constraints = {0: (0.0, None)}
        >>> LogCEI = LogConstrainedExpectedImprovement(model, 0.2, 1, constraints)
        >>> cei = LogCEI(test_X)
    """

    _log: bool = True

    def __init__(
        self,
        model: Model,
        best_f: float | Tensor,
        objective_index: int,
        constraints: dict[int, tuple[float | None, float | None]],
        maximize: bool = True,
    ) -> None:
        r"""Analytic Log Constrained Expected Improvement.

        Args:
            model: A fitted single- or multi-output model.
            best_f: Either a scalar or a `b`-dim Tensor (batch mode) representing
                the best feasible function value observed so far (assumed noiseless).
            objective_index: The index of the objective.
            constraints: A dictionary of the form `{i: [lower, upper]}`, where
                `i` is the output index, and `lower` and `upper` are lower and upper
                bounds on that output (resp. interpreted as -Inf / Inf if None)
            maximize: If True, consider the problem a maximization problem.
        """
        super().__init__(model=model, allow_multi_output=True)
        self.posterior_transform = None
        self.maximize = maximize
        self.objective_index = objective_index
        self.register_buffer("best_f", torch.as_tensor(best_f))
        ConstrainedAnalyticAcquisitionFunctionMixin.__init__(self, constraints)

    @t_batch_mode_transform(expected_q=1)
    @average_over_ensemble_models
    def forward(self, X: Tensor) -> Tensor:
        r"""Evaluate Constrained Log Expected Improvement on the candidate set X.

        Args:
            X: A `(b) x 1 x d`-dim Tensor of `(b)` t-batches of `d`-dim design
                points each.

        Returns:
            A `(b)`-dim Tensor of Log Expected Improvement values at the given
            design points `X`.
        """
        means, sigmas = self._mean_and_sigma(X)  # (b) x 1 + (m = num constraints)
        ind = self.objective_index
        mean_obj, sigma_obj = means[..., ind], sigmas[..., ind]
        u = _scaled_improvement(mean_obj, sigma_obj, self.best_f, self.maximize)
        log_ei = _log_ei_helper(u) + sigma_obj.log()
        log_prob_feas = self._compute_log_prob_feas(means=means, sigmas=sigmas)
        return log_ei + log_prob_feas


class LogProbabilityOfFeasibility(
    AnalyticAcquisitionFunction, ConstrainedAnalyticAcquisitionFunctionMixin
):
    r"""Log Probability of Feasbility.

    Computes the logarithm of the analytic probability of feasibility for a Normal
    posterior distribution weighted by a probability of feasibility. The objective and
    constraints are assumed to be independent and have Gaussian posterior
    distributions. Only supports non-batch mode (i.e. `q=1`). The model should be
    multi-outcome, with the index of the objective and constraints passed to
    the constructor.

    See [Ament2023logei]_ for details. Formally,

    `LogPF(x) = Sum_i log(P(y_i \in [lower_i, upper_i]))`,

    where `y_i ~ constraint_i(x)` and `lower_i`, `upper_i` are the lower and
    upper bounds for the i-th constraint, respectively.

    Example:
        # example where the 0th output has a non-negativity constraint
        >>> model = SingleTaskGP(train_X, train_Y)
        >>> constraints = {0: (0.0, None)}
        >>> LogPOF = LogProbabilityOfFeasibility(model, constraints)
        >>> cei = LogPF(test_X)
    """

    _log: bool = True

    def __init__(
        self, model: Model, constraints: dict[int, tuple[float | None, float | None]]
    ) -> None:
        r"""Analytic Log Probability of Feasibility.

        Args:
            model: A fitted single- or multi-output model.
            constraints: A dictionary of the form `{i: [lower, upper]}`, where
                `i` is the output index, and `lower` and `upper` are lower and upper
                bounds on that output (resp. interpreted as -Inf / Inf if None)
        """
        super().__init__(model=model, allow_multi_output=True)
        self.posterior_transform = None
        ConstrainedAnalyticAcquisitionFunctionMixin.__init__(self, constraints)

    @t_batch_mode_transform(expected_q=1)
    @average_over_ensemble_models
    def forward(self, X: Tensor) -> Tensor:
        r"""Evaluate Constrained Log Probability of Feasibility on the candidate set X.

        Args:
            X: A `(b) x 1 x d`-dim Tensor of `(b)` t-batches of `d`-dim design
                points each.

        Returns:
            A `(b)`-dim Tensor of Log Probability of Feasibility values at the given
            design points `X`.
        """
        means, sigmas = self._mean_and_sigma(X)  # (b) x 1 + (m = num constraints)
        return self._compute_log_prob_feas(means=means, sigmas=sigmas)


class ConstrainedExpectedImprovement(
    AnalyticAcquisitionFunction, ConstrainedAnalyticAcquisitionFunctionMixin
):
    r"""Constrained Expected Improvement (feasibility-weighted).

    Computes the analytic expected improvement for a Normal posterior
    distribution, weighted by a probability of feasibility. The objective and
    constraints are assumed to be independent and have Gaussian posterior
    distributions. Only supports non-batch mode (i.e. `q=1`). The model should be
    multi-outcome, with the index of the objective and constraints passed to
    the constructor.

    `Constrained_EI(x) = EI(x) * Product_i P(y_i \in [lower_i, upper_i])`,
    where `y_i ~ constraint_i(x)` and `lower_i`, `upper_i` are the lower and
    upper bounds for the i-th constraint, respectively.

    Example:
        # example where the 0th output has a non-negativity constraint and
        # 1st output is the objective
        >>> model = SingleTaskGP(train_X, train_Y)
        >>> constraints = {0: (0.0, None)}
        >>> cEI = ConstrainedExpectedImprovement(model, 0.2, 1, constraints)
        >>> cei = cEI(test_X)

    NOTE: It is strongly recommended to use LogConstrainedExpectedImprovement instead
    of regular CEI, as it can lead to substantially improved BO performance through
    improved numerics. See https://arxiv.org/abs/2310.20708 for details.
    """

    def __init__(
        self,
        model: Model,
        best_f: float | Tensor,
        objective_index: int,
        constraints: dict[int, tuple[float | None, float | None]],
        maximize: bool = True,
    ) -> None:
        r"""Analytic Constrained Expected Improvement.

        Args:
            model: A fitted single- or multi-output model.
            best_f: Either a scalar or a `b`-dim Tensor (batch mode) representing
                the best feasible function value observed so far (assumed noiseless).
            objective_index: The index of the objective.
            constraints: A dictionary of the form `{i: [lower, upper]}`, where
                `i` is the output index, and `lower` and `upper` are lower and upper
                bounds on that output (resp. interpreted as -Inf / Inf if None)
            maximize: If True, consider the problem a maximization problem.
        """
        legacy_ei_numerics_warning(legacy_name=type(self).__name__)
        super().__init__(model=model, allow_multi_output=True)
        self.posterior_transform = None
        self.maximize = maximize
        self.objective_index = objective_index
        self.register_buffer("best_f", torch.as_tensor(best_f))
        ConstrainedAnalyticAcquisitionFunctionMixin.__init__(self, constraints)

    @t_batch_mode_transform(expected_q=1)
    @average_over_ensemble_models
    def forward(self, X: Tensor) -> Tensor:
        r"""Evaluate Constrained Expected Improvement on the candidate set X.

        Args:
            X: A `(b) x 1 x d`-dim Tensor of `(b)` t-batches of `d`-dim design
                points each.

        Returns:
            A `(b)`-dim Tensor of Expected Improvement values at the given
            design points `X`.
        """
        means, sigmas = self._mean_and_sigma(X)  # (b) x 1 + (m = num constraints)
        ind = self.objective_index
        mean_obj, sigma_obj = means[..., ind], sigmas[..., ind]
        u = _scaled_improvement(mean_obj, sigma_obj, self.best_f, self.maximize)
        ei = sigma_obj * _ei_helper(u)
        log_prob_feas = self._compute_log_prob_feas(means=means, sigmas=sigmas)
        return ei.mul(log_prob_feas.exp())


class LogNoisyExpectedImprovement(AnalyticAcquisitionFunction):
    r"""Single-outcome Log Noisy Expected Improvement (via fantasies).

    This computes Log Noisy Expected Improvement by averaging over the Expected
    Improvement values of a number of fantasy models. Only supports the case
    `q=1`. Assumes that the posterior distribution of the model is Gaussian.
    The model must be single-outcome.

    See [Ament2023logei]_ for details. Formally,

    `LogNEI(x) = log(E(max(y - max Y_base), 0))), (y, Y_base) ~ f((x, X_base))`,

    where `X_base` are previously observed points.

    Note: This acquisition function currently relies on using a SingleTaskGP
    with known observation noise. In other words, `train_Yvar` must be passed
    to the model. (required for noiseless fantasies).

    Example:
        >>> model = SingleTaskGP(train_X, train_Y, train_Yvar=train_Yvar)
        >>> LogNEI = LogNoisyExpectedImprovement(model, train_X)
        >>> nei = LogNEI(test_X)
    """

    _log: bool = True

    def __init__(
        self,
        model: GPyTorchModel,
        X_observed: Tensor,
        num_fantasies: int = 20,
        maximize: bool = True,
        posterior_transform: PosteriorTransform | None = None,
    ) -> None:
        r"""Single-outcome Noisy Log Expected Improvement (via fantasies).

        Args:
            model: A fitted single-outcome model. Only `SingleTaskGP` models with
                known observation noise are currently supported.
            X_observed: A `n x d` Tensor of observed points that are likely to
                be the best observed points so far.
            num_fantasies: The number of fantasies to generate. The higher this
                number the more accurate the model (at the expense of model
                complexity and performance).
            maximize: If True, consider the problem a maximization problem.
        """
        _check_noisy_ei_model(model=model)
        # Sample fantasies.
        from botorch.sampling.normal import SobolQMCNormalSampler

        # Drop gradients from model.posterior if X_observed does not require gradients
        # as otherwise, gradients of the GP's kernel's hyper-parameters are tracked
        # through the rsample_from_base_sample method of GPyTorchPosterior. These
        # gradients are usually only required w.r.t. the marginal likelihood.
        with nullcontext() if X_observed.requires_grad else torch.no_grad():
            posterior = model.posterior(X=X_observed)
        sampler = SobolQMCNormalSampler(sample_shape=torch.Size([num_fantasies]))
        Y_fantasized = sampler(posterior).squeeze(-1)
        batch_X_observed = X_observed.expand(num_fantasies, *X_observed.shape)
        # The fantasy model will operate in batch mode
        fantasy_model = _get_noiseless_fantasy_model(
            model=model, batch_X_observed=batch_X_observed, Y_fantasized=Y_fantasized
        )
        super().__init__(model=fantasy_model, posterior_transform=posterior_transform)
        best_f, _ = Y_fantasized.max(dim=-1) if maximize else Y_fantasized.min(dim=-1)
        self.best_f, self.maximize = best_f, maximize

    @t_batch_mode_transform(expected_q=1)
    @average_over_ensemble_models
    def forward(self, X: Tensor) -> Tensor:
        r"""Evaluate logarithm of the mean Expected Improvement on the candidate set X.

        Args:
            X: A `b1 x ... bk x 1 x d`-dim batched tensor of `d`-dim design points.

        Returns:
            A `b1 x ... bk`-dim tensor of Log Noisy Expected Improvement values at
            the given design points `X`.
        """
        # add batch dimension for broadcasting to fantasy models
        # (b1 x ... x bk) x num_fantasies x 1
        mean, sigma = self._mean_and_sigma(X.unsqueeze(-3))
        mean, sigma = mean.squeeze(-1), sigma.squeeze(-1)
        u = _scaled_improvement(mean, sigma, self.best_f, self.maximize)
        log_ei = _log_ei_helper(u) + sigma.log()
        # this is mathematically - though not numerically - equivalent to log(mean(ei))
        return logmeanexp(log_ei, dim=-1)


class NoisyExpectedImprovement(ExpectedImprovement):
    r"""Single-outcome Noisy Expected Improvement (via fantasies).

    This computes Noisy Expected Improvement by averaging over the Expected
    Improvement values of a number of fantasy models. Only supports the case
    `q=1`. Assumes that the posterior distribution of the model is Gaussian.
    The model must be single-outcome.

    `NEI(x) = E(max(y - max Y_baseline), 0)), (y, Y_baseline) ~ f((x, X_baseline))`,
    where `X_baseline` are previously observed points.

    Note: This acquisition function currently relies on using a SingleTaskGP
    with known observation noise. In other words, `train_Yvar` must be passed
    to the model. (required for noiseless fantasies).

    Example:
        >>> model = SingleTaskGP(train_X, train_Y, train_Yvar=train_Yvar)
        >>> NEI = NoisyExpectedImprovement(model, train_X)
        >>> nei = NEI(test_X)

    NOTE: It is strongly recommended to use LogNoisyExpectedImprovement instead
    of regular NEI, as it can lead to substantially improved BO performance through
    improved numerics. See https://arxiv.org/abs/2310.20708 for details.
    """

    def __init__(
        self,
        model: GPyTorchModel,
        X_observed: Tensor,
        num_fantasies: int = 20,
        maximize: bool = True,
    ) -> None:
        r"""Single-outcome Noisy Expected Improvement (via fantasies).

        Args:
            model: A fitted single-outcome model. Only `SingleTaskGP` models with
                known observation noise are currently supported.
            X_observed: A `n x d` Tensor of observed points that are likely to
                be the best observed points so far.
            num_fantasies: The number of fantasies to generate. The higher this
                number the more accurate the model (at the expense of model
                complexity and performance).
            maximize: If True, consider the problem a maximization problem.
        """
        _check_noisy_ei_model(model=model)
        legacy_ei_numerics_warning(legacy_name=type(self).__name__)
        # Sample fantasies.
        from botorch.sampling.normal import SobolQMCNormalSampler

        # Drop gradients from model.posterior if X_observed does not require gradients
        # as otherwise, gradients of the GP's kernel's hyper-parameters are tracked
        # through the rsample_from_base_sample method of GPyTorchPosterior. These
        # gradients are usually only required w.r.t. the marginal likelihood.
        with nullcontext() if X_observed.requires_grad else torch.no_grad():
            posterior = model.posterior(X=X_observed)
        sampler = SobolQMCNormalSampler(sample_shape=torch.Size([num_fantasies]))
        Y_fantasized = sampler(posterior).squeeze(-1)
        batch_X_observed = X_observed.expand(num_fantasies, *X_observed.shape)
        # The fantasy model will operate in batch mode
        fantasy_model = _get_noiseless_fantasy_model(
            model=model, batch_X_observed=batch_X_observed, Y_fantasized=Y_fantasized
        )
        best_f, _ = Y_fantasized.max(dim=-1) if maximize else Y_fantasized.min(dim=-1)
        super().__init__(model=fantasy_model, best_f=best_f, maximize=maximize)

    @t_batch_mode_transform(expected_q=1)
    @average_over_ensemble_models
    def forward(self, X: Tensor) -> Tensor:
        r"""Evaluate Expected Improvement on the candidate set X.

        Args:
            X: A `(b1 x ... bk) x 1 x d`-dim batched tensor of `d`-dim design points.

        Returns:
            A `(b1 x ... bk)`-dim tensor of Noisy Expected Improvement values at
            the given design points `X`.
        """
        # add batch dimension for broadcasting to fantasy models
        # (b1 x ... x bk) x num_fantasies x 1
        mean, sigma = self._mean_and_sigma(X.unsqueeze(-3))  # (b1 x ... x bk) x m1 x 1
        mean, sigma = mean.squeeze(-1), sigma.squeeze(-1)
        u = _scaled_improvement(mean, sigma, self.best_f, self.maximize)
        return (sigma * _ei_helper(u)).mean(dim=-1)


class UpperConfidenceBound(AnalyticAcquisitionFunction):
    r"""Single-outcome Upper Confidence Bound (UCB).

    Analytic upper confidence bound that comprises of the posterior mean plus an
    additional term: the posterior standard deviation weighted by a trade-off
    parameter, `beta`. Only supports the case of `q=1` (i.e. greedy, non-batch
    selection of design points). The model must be single-outcome.

    `UCB(x) = mu(x) + sqrt(beta) * sigma(x)`, where `mu` and `sigma` are the
    posterior mean and standard deviation, respectively.

    Example:
        >>> model = SingleTaskGP(train_X, train_Y)
        >>> UCB = UpperConfidenceBound(model, beta=0.2)
        >>> ucb = UCB(test_X)
    """

    def __init__(
        self,
        model: Model,
        beta: float | Tensor,
        posterior_transform: PosteriorTransform | None = None,
        maximize: bool = True,
    ) -> None:
        r"""Single-outcome Upper Confidence Bound.

        Args:
            model: A fitted single-outcome GP model (must be in batch mode if
                candidate sets X will be)
            beta: Either a scalar or a one-dim tensor with `b` elements (batch mode)
                representing the trade-off parameter between mean and covariance
            posterior_transform: A PosteriorTransform. If using a multi-output model,
                a PosteriorTransform that transforms the multi-output posterior into a
                single-output posterior is required.
            maximize: If True, consider the problem a maximization problem.
        """
        super().__init__(model=model, posterior_transform=posterior_transform)
        self.register_buffer("beta", torch.as_tensor(beta))
        self.maximize = maximize

    @t_batch_mode_transform(expected_q=1)
    @average_over_ensemble_models
    def forward(self, X: Tensor) -> Tensor:
        r"""Evaluate the Upper Confidence Bound on the candidate set X.

        Args:
            X: A `(b1 x ... bk) x 1 x d`-dim batched tensor of `d`-dim design points.

        Returns:
            A `(b1 x ... bk)`-dim tensor of Upper Confidence Bound values at the
            given design points `X`.
        """
        mean, sigma = self._mean_and_sigma(X)  # (b1 x ... x bk) x 1
        ucb = (mean if self.maximize else -mean) + self.beta.sqrt() * sigma
        return ucb.squeeze(-1)


class PosteriorMean(AnalyticAcquisitionFunction):
    r"""Single-outcome Posterior Mean.

    Only supports the case of q=1. Requires the model's posterior to have a
    `mean` property. The model must be single-outcome.

    Example:
        >>> model = SingleTaskGP(train_X, train_Y)
        >>> PM = PosteriorMean(model)
        >>> pm = PM(test_X)
    """

    def __init__(
        self,
        model: Model,
        posterior_transform: PosteriorTransform | None = None,
        maximize: bool = True,
    ) -> None:
        r"""Single-outcome Posterior Mean.

        Args:
            model: A fitted single-outcome GP model (must be in batch mode if
                candidate sets X will be)
            posterior_transform: A PosteriorTransform. If using a multi-output model,
                a PosteriorTransform that transforms the multi-output posterior into a
                single-output posterior is required.
            maximize: If True, consider the problem a maximization problem. Note
                that if `maximize=False`, the posterior mean is negated. As a
                consequence `optimize_acqf(PosteriorMean(gp, maximize=False))`
                actually returns -1 * minimum of the posterior mean.
        """
        super().__init__(model=model, posterior_transform=posterior_transform)
        self.maximize = maximize

    @t_batch_mode_transform(expected_q=1)
    @average_over_ensemble_models
    def forward(self, X: Tensor) -> Tensor:
        r"""Evaluate the posterior mean on the candidate set X.

        Args:
            X: A `(b1 x ... bk) x 1 x d`-dim batched tensor of `d`-dim design points.

        Returns:
            A `(b1 x ... bk)`-dim tensor of Posterior Mean values at the
            given design points `X`.
        """
        mean, _ = self._mean_and_sigma(X, compute_sigma=False)  # (b1 x ... x bk) x 1
        if not self.maximize:
            mean = -mean
        return mean.squeeze(-1)


class ScalarizedPosteriorMean(AnalyticAcquisitionFunction):
    r"""Scalarized Posterior Mean.

    This acquisition function returns a scalarized (across the q-batch)
    posterior mean given a vector of weights.
    """

    def __init__(
        self,
        model: Model,
        weights: Tensor,
        posterior_transform: PosteriorTransform | None = None,
    ) -> None:
        r"""Scalarized Posterior Mean.

        Args:
            model: A fitted single-outcome model.
            weights: A tensor of shape `q` for scalarization. In order to minimize
                the scalarized posterior mean, pass -weights.
            posterior_transform: A PosteriorTransform. If using a multi-output model,
                a PosteriorTransform that transforms the multi-output posterior into a
                single-output posterior is required.
        """
        super().__init__(model=model, posterior_transform=posterior_transform)
        self.register_buffer("weights", weights)

    @t_batch_mode_transform()
    @average_over_ensemble_models
    def forward(self, X: Tensor) -> Tensor:
        r"""Evaluate the scalarized posterior mean on the candidate set X.

        Args:
            X: A `(b1 x ... x bk) x q x d`-dim Tensor of `(b1 x ... x bk)`
                t-batches of `d`-dim design points each.

        Returns:
            A `(b1 x ... x bk)`-dim Tensor of Posterior Mean values at the given
            design points `X`.
        """
        # (b1 x ... x bk) x q x 1
        mean, _ = self._mean_and_sigma(X, compute_sigma=False)
        return mean.squeeze(-1) @ self.weights


class PosteriorStandardDeviation(AnalyticAcquisitionFunction):
    r"""Single-outcome Posterior Standard Deviation.

    An acquisition function for pure exploration.
    Only supports the case of q=1. Requires the model's posterior to have
    `mean` and `variance` properties. The model must be either single-outcome
    or combined with a `posterior_transform` to produce a single-output posterior.

    Example:
        >>> import torch
        >>> from botorch.models.gp_regression import SingleTaskGP
        >>> from botorch.models.transforms.input import Normalize
        >>> from botorch.models.transforms.outcome import Standardize
        >>>
        >>> # Set up a model
        >>> train_X = torch.rand(20, 2, dtype=torch.float64)
        >>> train_Y = torch.sin(train_X).sum(dim=1, keepdim=True)
        >>> model = SingleTaskGP(
        ...     train_X, train_Y, outcome_transform=Standardize(m=1),
        ...     input_transform=Normalize(d=2),
        ... )
        >>> # Now set up the acquisition function
        >>> PSTD = PosteriorStandardDeviation(model)
        >>> test_X = torch.zeros((1, 2), dtype=torch.float64)
        >>> std = PSTD(test_X)
        >>> std.item()
        0.16341639895667773
    """

    def __init__(
        self,
        model: Model,
        posterior_transform: PosteriorTransform | None = None,
        maximize: bool = True,
    ) -> None:
        r"""Single-outcome Posterior Mean.

        Args:
            model: A fitted single-outcome GP model (must be in batch mode if
                candidate sets X will be)
            posterior_transform: A PosteriorTransform. If using a multi-output model,
                a PosteriorTransform that transforms the multi-output posterior into a
                single-output posterior is required.
            maximize: If True, consider the problem a maximization problem. Note
                that if `maximize=False`, the posterior standard deviation is negated.
                As a consequence,
                `optimize_acqf(PosteriorStandardDeviation(gp, maximize=False))`
                actually returns -1 * minimum of the posterior standard deviation.
        """
        super().__init__(model=model, posterior_transform=posterior_transform)
        self.maximize = maximize

    @t_batch_mode_transform(expected_q=1)
    @average_over_ensemble_models
    def forward(self, X: Tensor) -> Tensor:
        r"""Evaluate the posterior standard deviation on the candidate set X.

        Args:
            X: A `(b1 x ... bk) x 1 x d`-dim batched tensor of `d`-dim design points.

        Returns:
            A `(b1 x ... bk)`-dim tensor of Posterior Mean values at the
            given design points `X`.
        """
        _, std = self._mean_and_sigma(X)
        if not self.maximize:
            std = -std
        return std.view(X.shape[:-2])


# --------------- Helper functions for analytic acquisition functions. ---------------


def _scaled_improvement(
    mean: Tensor, sigma: Tensor, best_f: Tensor, maximize: bool
) -> Tensor:
    """Returns `u = (mean - best_f) / sigma`, -u if maximize == True."""
    u = (mean - best_f) / sigma
    return u if maximize else -u


def _ei_helper(u: Tensor) -> Tensor:
    """Computes phi(u) + u * Phi(u), where phi and Phi are the standard normal
    pdf and cdf, respectively. This is used to compute Expected Improvement.
    """
    return phi(u) + u * Phi(u)


def _log_ei_helper(u: Tensor) -> Tensor:
    """Accurately computes log(phi(u) + u * Phi(u)) in a differentiable manner for u in
    [-10^100, 10^100] in double precision, and [-10^20, 10^20] in single precision.
    Beyond these intervals, a basic squaring of u can lead to floating point overflow.
    In contrast, the implementation in _ei_helper only yields usable gradients down to
    u ~ -10. As a consequence, _log_ei_helper improves the range of inputs for which a
    backward pass yields usable gradients by many orders of magnitude.
    """
    if not (u.dtype == torch.float32 or u.dtype == torch.float64):
        raise TypeError(
            f"LogExpectedImprovement only supports torch.float32 and torch.float64 "
            f"dtypes, but received {u.dtype=}."
        )
    # The function has two branching decisions. The first is u < bound, and in this
    # case, just taking the logarithm of the naive _ei_helper implementation works.
    bound = -1
    u_upper = u.masked_fill(u < bound, bound)  # mask u to avoid NaNs in gradients
    log_ei_upper = _ei_helper(u_upper).log()

    # When u <= bound, we need to be more careful and rearrange the EI formula as
    # log(phi(u)) + log(1 - exp(w)), where w = log(abs(u) * Phi(u) / phi(u)).
    # To this end, a second branch is necessary, depending on whether or not u is
    # smaller than approximately the negative inverse square root of the machine
    # precision. Below this point, numerical issues in computing log(1 - exp(w)) occur
    # as w approaches zero from below, even though the relative contribution to log_ei
    # vanishes in machine precision at that point.
    neg_inv_sqrt_eps = -1e6 if u.dtype == torch.float64 else -1e3

    # mask u for to avoid NaNs in gradients in first and second branch
    u_lower = u.masked_fill(u > bound, bound)
    u_eps = u_lower.masked_fill(u < neg_inv_sqrt_eps, neg_inv_sqrt_eps)
    # compute the logarithm of abs(u) * Phi(u) / phi(u) for moderately large negative u
    w = _log_abs_u_Phi_div_phi(u_eps)

    # 1) Now, we use a special implementation of log(1 - exp(w)) for moderately
    # large negative numbers, and
    # 2) capture the leading order of log(1 - exp(w)) for very large negative numbers.
    # The second special case is technically only required for single precision numbers
    # but does "the right thing" regardless.
    log_ei_lower = log_phi(u) + (
        torch.where(
            u > neg_inv_sqrt_eps,
            log1mexp(w),
            # The contribution of the next term relative to log_phi vanishes when
            # w_lower << eps but captures the leading order of the log1mexp term.
            -2 * u_lower.abs().log(),
        )
    )
    return torch.where(u > bound, log_ei_upper, log_ei_lower)


def _log_abs_u_Phi_div_phi(u: Tensor) -> Tensor:
    """Computes log(abs(u) * Phi(u) / phi(u)), where phi and Phi are the normal pdf
    and cdf, respectively. The function is valid for u < 0.

    NOTE: In single precision arithmetic, the function becomes numerically unstable for
    u < -1e3. For this reason, a second branch in _log_ei_helper is necessary to handle
    this regime, where this function approaches -abs(u)^-2 asymptotically.

    The implementation is based on the following implementation of the logarithm of
    the scaled complementary error function (i.e. erfcx). Since we only require the
    positive branch for _log_ei_helper, _log_abs_u_Phi_div_phi does not have a branch,
    but is only valid for u < 0 (so that _neg_inv_sqrt2 * u > 0).

        def logerfcx(x: Tensor) -> Tensor:
            return torch.where(
                x < 0,
                torch.erfc(x.masked_fill(x > 0, 0)).log() + x**2,
                torch.special.erfcx(x.masked_fill(x < 0, 0)).log(),
        )

    Further, it is important for numerical accuracy to move u.abs() into the
    logarithm, rather than adding u.abs().log() to logerfcx. This is the reason
    for the rather complex name of this function: _log_abs_u_Phi_div_phi.
    """
    # get_constants_like allocates tensors with the appropriate dtype and device and
    # caches the result, which improves efficiency.
    a, b = get_constants_like(values=(_neg_inv_sqrt2, _log_sqrt_pi_div_2), ref=u)
    return torch.log(torch.special.erfcx(a * u) * u.abs()) + b


def _check_noisy_ei_model(model: GPyTorchModel) -> None:
    message = (
        "Only single-output `SingleTaskGP` models with known observation noise "
        "are currently supported for fantasy-based NEI & LogNEI."
    )
    if not isinstance(model, SingleTaskGP):
        raise UnsupportedError(f"{message} Model is not a `SingleTaskGP`.")
    if not isinstance(model.likelihood, FixedNoiseGaussianLikelihood):
        raise UnsupportedError(
            f"{message} Model likelihood is not a `FixedNoiseGaussianLikelihood`."
        )
    if model.num_outputs != 1:
        raise UnsupportedError(f"{message} Model has {model.num_outputs} outputs.")


def _get_noiseless_fantasy_model(
    model: SingleTaskGP, batch_X_observed: Tensor, Y_fantasized: Tensor
) -> SingleTaskGP:
    r"""Construct a fantasy model from a fitted model and provided fantasies.

    The fantasy model uses the hyperparameters from the original fitted model and
    assumes the fantasies are noiseless.

    Args:
        model: A fitted SingleTaskGP with known observation noise.
        batch_X_observed: A `b x n x d` tensor of inputs where `b` is the number of
            fantasies.
        Y_fantasized: A `b x n` tensor of fantasized targets where `b` is the number of
            fantasies.

    Returns:
        The fantasy model.
    """
    # initialize a copy of SingleTaskGP on the original training inputs
    # this makes SingleTaskGP a non-batch GP, so that the same hyperparameters
    # are used across all batches (by default, a GP with batched training data
    # uses independent hyperparameters for each batch).

    # We don't want to use the true `outcome_transform` and `input_transform` here
    # since the data being passed has already been transformed. We thus pass `None`
    # and will instead set them afterwards.
    fantasy_model = SingleTaskGP(
        train_X=model.train_inputs[0],
        train_Y=model.train_targets.unsqueeze(-1),
        train_Yvar=model.likelihood.noise_covar.noise.unsqueeze(-1),
        covar_module=deepcopy(model.covar_module),
        mean_module=deepcopy(model.mean_module),
        outcome_transform=None,
        input_transform=None,
    )

    Yvar = torch.full_like(Y_fantasized, 1e-7)

    # Set the outcome and input transforms of the fantasy model.
    # The transforms should already be in eval mode but just set them to be sure
    outcome_transform = getattr(model, "outcome_transform", None)
    if outcome_transform is not None:
        outcome_transform = deepcopy(outcome_transform).eval()
        fantasy_model.outcome_transform = outcome_transform
        # Need to transform the outcome just as in the SingleTaskGP constructor.
        # Need to unsqueeze for BoTorch and then squeeze again for GPyTorch.
        # Not transforming Yvar because 1e-7 is already close to 0 and it is a
        # relative, not absolute, value.
        Y_fantasized, _ = outcome_transform(
            Y_fantasized.unsqueeze(-1), Yvar.unsqueeze(-1), X=batch_X_observed
        )
        Y_fantasized = Y_fantasized.squeeze(-1)
    input_transform = getattr(model, "input_transform", None)
    if input_transform is not None:
        fantasy_model.input_transform = deepcopy(input_transform).eval()

    # update training inputs/targets to be batch mode fantasies
    fantasy_model.set_train_data(
        inputs=batch_X_observed, targets=Y_fantasized, strict=False
    )
    # use noiseless fantasies
    fantasy_model.likelihood.noise_covar.noise = Yvar

    return fantasy_model



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import torch
from botorch.utils.testing import BotorchTestCase
from botorch_community.utils.stat_dist import mvn_hellinger_distance, mvn_kl_divergence


class TestStatDist(BotorchTestCase):
    def test_mvn_kl_divergence(self):
        q_shapes = [(1,), (5,)]
        mvn1_batch_shapes = [(4,), (4, 5), (4, 5), (4, 5, 7), (1, 5, 9, 2), (1, 2, 3)]
        mvn2_batch_shapes = [(4,), (4, 5), (1, 5), (4, 5, 1), (7, 5, 9, 2), (3, 2, 1)]

        for mvn1_shape, mvn2_shape in zip(mvn1_batch_shapes, mvn2_batch_shapes):
            for qdim in q_shapes:
                dist1_mean = torch.zeros(mvn1_shape + qdim + (1,))
                dist2_mean = torch.zeros(mvn2_shape + qdim + (1,))

                dist1_cov = torch.diag_embed(
                    torch.ones(dist1_mean.shape[:-1]), dim1=-2, dim2=-1
                )
                dist2_cov = torch.diag_embed(
                    torch.ones(dist2_mean.shape[:-1]), dim1=-2, dim2=-1
                )

                # creating a diagonal matrix of covariances
                res = mvn_kl_divergence(dist1_mean, dist2_mean, dist1_cov, dist2_cov)

                correct_shape = torch.broadcast_shapes(mvn1_shape, mvn2_shape) + (1,)
                self.assertEqual(res.shape, correct_shape)
                self.assertTrue(torch.all(res == 0))

        perturb_mean = dist1_mean + 1
        perturb_cov = dist2_cov + 0.01
        permean_res = mvn_kl_divergence(perturb_mean, dist2_mean, dist1_cov, dist2_cov)
        percov_res = mvn_kl_divergence(dist1_mean, dist2_mean, dist1_cov, perturb_cov)
        self.assertTrue(torch.all(permean_res > 0))
        self.assertTrue(torch.all(percov_res > 0))

    def test_mvn_hellinger_distance(self):
        q_shapes = [(1,), (5,)]
        mvn1_batch_shapes = [(4,), (4, 5), (4, 5), (4, 5, 7), (1, 5, 9, 2), (1, 2, 3)]
        mvn2_batch_shapes = [(4,), (4, 5), (1, 5), (4, 5, 1), (7, 5, 9, 2), (3, 2, 1)]

        for mvn1_shape, mvn2_shape in zip(mvn1_batch_shapes, mvn2_batch_shapes):
            for qdim in q_shapes:
                dist1_mean = torch.zeros(mvn1_shape + qdim + (1,))
                dist2_mean = torch.zeros(mvn2_shape + qdim + (1,))

                dist1_cov = torch.diag_embed(
                    torch.ones(dist1_mean.shape[:-1]), dim1=-2, dim2=-1
                )
                dist2_cov = torch.diag_embed(
                    torch.ones(dist2_mean.shape[:-1]), dim1=-2, dim2=-1
                )

                # creating a diagonal matrix of covariances
                res = mvn_hellinger_distance(
                    dist1_mean, dist2_mean, dist1_cov, dist2_cov
                )

                correct_shape = torch.broadcast_shapes(mvn1_shape, mvn2_shape) + (1,)
                self.assertEqual(res.shape, correct_shape)
                self.assertTrue(torch.all(res == 0))

        perturb_mean = dist1_mean + 1
        perturb_cov = dist2_cov + 0.01
        permean_res = mvn_hellinger_distance(
            perturb_mean, dist2_mean, dist1_cov, dist2_cov
        )
        percov_res = mvn_hellinger_distance(
            dist1_mean, dist2_mean, dist1_cov, perturb_cov
        )
        self.assertTrue(torch.all(permean_res > 0))
        self.assertTrue(torch.all(percov_res > 0))



import unittest

import torch
from botorch.models.transforms.input import Normalize
from botorch.posteriors import GPyTorchPosterior
from botorch_community.models.np_regression import NeuralProcessModel

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")


class Identity:
    def __call__(self, posterior):
        return posterior


class TestNeuralProcessModel(unittest.TestCase):
    def initialize(self):
        self.r_hidden_dims = [16, 16]
        self.z_hidden_dims = [32, 32]
        self.decoder_hidden_dims = [16, 16]
        self.x_dim = 2
        self.y_dim = 1
        self.r_dim = 8
        self.z_dim = 8
        self.n_context = 20
        self.model = NeuralProcessModel(
            torch.rand(100, self.x_dim),
            torch.rand(100, self.y_dim),
            self.r_hidden_dims,
            self.z_hidden_dims,
            self.decoder_hidden_dims,
            self.x_dim,
            self.y_dim,
            self.r_dim,
            self.z_dim,
            self.n_context,
        )

    def test_r_encoder(self):
        self.initialize()
        input = torch.rand(100, self.x_dim + self.y_dim)
        output = self.model.r_encoder(input)
        self.assertEqual(output.shape, (100, self.r_dim))
        self.assertTrue(torch.is_tensor(output))

    def test_z_encoder(self):
        self.initialize()
        input = torch.rand(100, self.r_dim)
        mean, logvar = self.model.z_encoder(input)
        self.assertEqual(mean.shape, (100, self.z_dim))
        self.assertEqual(logvar.shape, (100, self.z_dim))
        self.assertTrue(torch.is_tensor(mean))
        self.assertTrue(torch.is_tensor(logvar))

    def test_decoder(self):
        self.initialize()
        x_pred = torch.rand(100, self.x_dim)
        z = torch.rand(self.z_dim)
        output = self.model.decoder(x_pred, z)
        self.assertEqual(output.shape, (100, self.y_dim))
        self.assertTrue(torch.is_tensor(output))

    def test_sample_z(self):
        self.initialize()
        mu = torch.rand(self.z_dim)
        logvar = torch.rand(self.z_dim)
        samples = self.model.sample_z(mu, logvar, n=5)
        self.assertEqual(samples.shape, (5, self.z_dim))
        self.assertTrue(torch.is_tensor(samples))
        with self.assertRaises(ValueError):
            self.model.sample_z(mu, logvar, n=5, scaler=-1)

    def test_KLD_gaussian(self):
        self.initialize()
        self.model.z_mu_all = torch.rand(self.z_dim)
        self.model.z_logvar_all = torch.rand(self.z_dim)
        self.model.z_mu_context = torch.rand(self.z_dim)
        self.model.z_logvar_context = torch.rand(self.z_dim)
        kld = self.model.KLD_gaussian()
        self.assertGreaterEqual(kld.item(), 0)
        self.assertTrue(torch.is_tensor(kld))
        with self.assertRaises(ValueError):
            self.model.KLD_gaussian(scaler=-1)

    def test_data_to_z_params(self):
        self.initialize()
        mu, logvar = self.model.data_to_z_params(self.model.train_X, self.model.train_Y)
        self.assertEqual(mu.shape, (self.z_dim,))
        self.assertEqual(logvar.shape, (self.z_dim,))
        self.assertTrue(torch.is_tensor(mu))
        self.assertTrue(torch.is_tensor(logvar))

    def test_forward(self):
        self.initialize()
        output = self.model(self.model.train_X, self.model.train_Y)
        self.assertEqual(output.loc.shape, (80, self.y_dim))

    def test_random_split_context_target(self):
        self.initialize()
        x_c, y_c, x_t, y_t = self.model.random_split_context_target(
            self.model.train_X[:, 0], self.model.train_Y, self.model.n_context
        )
        self.assertEqual(x_c.shape[0], 20)
        self.assertEqual(y_c.shape[0], 20)
        self.assertEqual(x_t.shape[0], 80)
        self.assertEqual(y_t.shape[0], 80)

    def test_posterior(self):
        self.initialize()
        self.model(self.model.train_X, self.model.train_Y)
        identity_posterior = self.model.posterior(
            self.model.train_X, observation_noise=True, posterior_transform=Identity()
        )
        posterior = self.model.posterior(self.model.train_X, observation_noise=True)
        self.assertIsInstance(identity_posterior, GPyTorchPosterior)
        self.assertIsInstance(posterior, GPyTorchPosterior)
        mvn = posterior.mvn
        self.assertEqual(mvn.covariance_matrix.size(), (100, 100, 100))

    def test_transform_inputs(self):
        self.initialize()
        X = torch.rand(5, 3)
        self.assertTrue(torch.equal(self.model.transform_inputs(X), X.to(device)))
        self.assertFalse(
            torch.equal(
                self.model.transform_inputs(X, input_transform=Normalize(d=3)),
                X.to(device),
            )
        )


if __name__ == "__main__":
    unittest.main()
